import os
import hashlib
import csv
import sys
from fhir_lib import get_fhir_pseudonym

def read_csv_table(table_path):
    with open(table_path) as f:
        reader = csv.reader(f)
        pseudonyms_dict = {str(rows[0]):str(rows[1]) for rows in reader}
        f.close()
    return pseudonyms_dict


def write_temp_table(ptid,pseudonym,temp_pse):
    to_write_line = 'ptid/{}={}\n'.format(ptid,pseudonym)
    with open(temp_pse, "w") as f:
        f.write(to_write_line)
        f.close()

def append_hash_table(table_path,ptid,pseudonym):
    with open(table_path, 'a+') as f:
        fieldnames = ['PatientID', 'Pseudonym']
    
        writer = csv.DictWriter(f, fieldnames=fieldnames)
        if os.path.getsize(table_path) ==0:
            writer.writeheader()
            
        writer.writerow({'PatientID': ptid,'Pseudonym':pseudonym})
        f.close()
def delete_temp_table(temp_pse):
    os.remove(temp_pse)

def get_correspondence(ptid,pseudonyms_dict,hash_table, params=None):
    print('getting correspondence', file=sys.stdout)

    if str(ptid) in pseudonyms_dict.keys():
        print('in pseudonyms table', file=sys.stdout)
        return pseudonyms_dict[ptid], False

    pseudonyms = get_fhir_pseudonym(ptid=ptid, params=params)
    if not pseudonyms is None:
        print('from fhir service', file=sys.stdout)
        return pseudonyms, True

    if os.path.isfile(hash_table):
        print('checking hashtable', file=sys.stdout)
        hash_dict = read_csv_table(hash_table)
        if str(ptid) in hash_dict.keys():
            return hash_dict[ptid], False
        else: 
            return hashlib.md5(str(ptid+'Y7t5').encode('utf-8')).hexdigest(), True
    else:
        return hashlib.md5(str(ptid+'Y7t5').encode('utf-8')).hexdigest(), True



import os
import pydicom
from .utils import read_csv_table, write_temp_table, append_hash_table, delete_temp_table, get_correspondence

MODULES = os.getenv('MODULES')
DATA = os.getenv('DATA')

CONFIG = os.path.join(os.path.dirname(os.path.realpath(__file__)),'config')

hash_table = os.path.join(CONFIG, 'hash_table.csv')
anony_script = os.path.join(CONFIG, 'dicom-anonymizer.script')
pseudonyms_table = os.path.join(CONFIG, 'psudonyms.csv')

temp_table = os.path.join(DATA, 'temp_table.properties')
anonymizer = os.path.join(os.path.dirname(os.path.realpath(__file__)),'DAT.jar')

def anonymize(path, params=None):
  series = []
  for r, _, f in os.walk(path):
    for file in f:
      series.append(os.path.join(r,file))

  ptid = pydicom.dcmread(series[0]).PatientID
  
  if os.path.isfile(pseudonyms_table):
    pseudonyms_dict = read_csv_table(pseudonyms_table)
      
  else:
    pseudonyms_dict = {}

  pseudonym, hashed = get_correspondence(ptid, pseudonyms_dict,hash_table, params=params)
  if hashed:
    append_hash_table(hash_table,ptid,pseudonym)

  write_temp_table(ptid,pseudonym,temp_table)
  
  anonymized_path = f"{path}_anon"

  an_path = os.path.join(anonymized_path)
  if not os.path.isdir(an_path):
    os.makedirs(an_path)
  
  command = "java -jar " + anonymizer +" -in " +'"' + path + '"' +" -out " + '"' + anonymized_path + '"'  + " -da " + '"' +anony_script + '"' + " -lut " + '"' +temp_table + '"' +" -v"

  os.system(command)
  os.system(f"rm -rf {path} && mv {anonymized_path} {path}")

  delete_temp_table(temp_table) 




import requests
import json
import time
import sys
from datetime import datetime

def get_config(params = None):
  config_dict = None

  if isinstance(params, dict):
    config_dict = params.get("fhir")

  return config_dict

def get_new_token(auth_server_url, client_id, client_secret):
  token_req_payload = {'grant_type': 'client_credentials'}
  token_response = requests.post(auth_server_url,data=token_req_payload, verify=False, allow_redirects=False,auth=(client_id, client_secret))
  time.sleep(1)
  if token_response.status_code !=200:
    print("Failed to obtain token from the OAuth 2.0 server", file=sys.stderr)
    return None
  else:
    tokens = json.loads(token_response.text)
    token = tokens['access_token']
    return token
    
def get_pseudononymized_bundle(auth_server_url, client_id, client_secret, interface_url, bundle):
  token = get_new_token(auth_server_url, client_id, client_secret)
  api_call_headers = {'Authorization': 'Bearer ' + token, 'Content-Type':'application/json'}
  response = requests.post(interface_url,headers=api_call_headers,data=json.dumps(bundle))
  time.sleep(2)
  if response.status_code !=401:
    token = get_new_token(auth_server_url, client_id, client_secret)
    response = requests.post(interface_url,headers=api_call_headers,data=json.dumps(bundle))
    time.sleep(5)
  if response.status_code ==200:
    pse_bundle = json.loads(response.text)
    return pse_bundle
  else:
    print('PSE server not responsive')
    raise Exception("Server Not reponding")
        
def get_pseudonyms_from_bundle(pse_bundle):
  pse_dict = None
  entries = pse_bundle.get("entry",None)
  if entries is not None:
    for entry in entries:
      if  entry['resource']['resourceType'] == 'Patient':
        pse_dict = entry['resource']['identifier'][0]['value']
  return pse_dict

def create_fhir_entry(resourceType, system, ID):
  entry = {
    "resource": {
      "resourceType": resourceType,
      "identifier": [
        {
          "use": "official",
          "system": system,
          "value": ID,
          "period": {
            "start": datetime.now().strftime("%Y-%m-%dT%H:%M:%S+01:00"),
            "end": datetime.now().strftime("%Y-%m-%dT%H:%M:%S+01:00")
          }
        }
      ]
    }
  }
  return entry

def create_fhir_bundle(entries):
  bundle = {
    "resourceType": "Bundle",
    "entry": entries
  }
  return bundle

# pat_entry = create_fhir_entry("Patient", identifier_system, 123456)
# accN_entry = create_fhir_entry("ImagingStudy", identifier_system, 123456789)
# bundle = create_fhir_bundle([pat_entry,accN_entry])

def get_fhir_pseudonym(ptid, params):
  config_dict = get_config(params)
  
  if config_dict is None:
    return None
  
  interface_url = config_dict.get('interface_url')
  auth_server_url = config_dict.get('auth_server_url')
  client_id = config_dict.get('client_id')
  client_secret = config_dict.get('client_secret')
  identifier_system = config_dict.get('identifier_system')

  if (interface_url is None) or (auth_server_url is None) or (client_id is None) or (client_secret) or (identifier_system is None):
    return None
  
  pat_entry = create_fhir_entry("Patient", identifier_system, ptid)
  
  bundle = create_fhir_bundle([pat_entry,])
  
  try:
    pse_bundle = get_pseudononymized_bundle(auth_server_url, client_id, client_secret, interface_url, bundle)
    pse_dict = get_pseudonyms_from_bundle(pse_bundle)
    return pse_dict
  except Exception as e:
    return None


import SimpleITK as sitk
import time, os, requests
import numpy as np
import os.path

def get_slope_and_intercept(image):
  arr = sitk.GetArrayFromImage(image)
  dtype = arr.dtype
  intercept = 0
  slope = 1
  changed = False

  if dtype in ["float32", "float64", "float", "int32", "int64", "int", "int16"]:
    min_val = 0
    max_val = 2**16
    img_min = np.min(arr)

    if img_min < min_val:
      intercept =  (min_val - img_min)

    img_max = np.max(arr) + intercept

    if img_max > max_val:
      slope = (max_val / img_max)

    changed = True

  return slope, intercept, dtype, changed

def write_slices(writer, series_tag_values, new_img, out_dir, i):
  image_slice = new_img[:, :, i]

  # Tags shared by the series.
  list(map(lambda tag_value: image_slice.SetMetaData(tag_value[0], tag_value[1]), series_tag_values))

  # Slice specific tags.
  #   Instance Creation Date
  image_slice.SetMetaData("0008|0012", time.strftime("%Y%m%d"))
  #   Instance Creation Time
  image_slice.SetMetaData("0008|0013", time.strftime("%H%M%S"))

  # Setting the type to CT so that the slice location is preserved and
  # the thickness is carried over.
  # image_slice.SetMetaData("0008|0060", "CT")

  # (0020, 0032) image position patient determines the 3D spacing between
  # slices.
  #   Image Position (Patient)
  image_slice.SetMetaData("0020|0032", '\\'.join(map(str, new_img.TransformIndexToPhysicalPoint((0, 0, i)))))
  #   Instance Number
  image_slice.SetMetaData("0020|0013", str(i + 1))

  # Write to the output directory and add the extension dcm, to force
  # writing in DICOM format.
  writer.SetFileName(os.path.join(out_dir, str(i + 1).zfill(5) + '.dcm'))
  writer.Execute(image_slice)

def nifti_to_dicom(filename, outpath, headers):

  reader = sitk.ImageFileReader()
  reader.SetFileName(filename)

  new_img = reader.Execute()
  slope, intercept, pixel_dtype, dtype_changed = get_slope_and_intercept(new_img)

  writer = sitk.ImageFileWriter()
  writer.KeepOriginalImageUIDOn()

  modification_time = time.strftime("%H%M%S")
  modification_date = time.strftime("%Y%m%d")

  direction = new_img.GetDirection()
  series_tag_values = headers
  series_tag_values = series_tag_values + [
    ("0008|0031", modification_time),  # Series Time
    ("0008|0021", modification_date),  # Series Date
    ('0020|0011', modification_date + modification_time), # Series Number
    ("0008|0008", "DERIVED\\SECONDARY"),  # Image Type
    ("0020|000e", "1.2.826.0.1.3680043.2.1125."
      + modification_date + ".1" + modification_time),  # Series Instance UID
    ("0020|0037", '\\'.join(map(str, (direction[0], direction[3], direction[6],
                                      direction[1], direction[4],
                                      direction[7])))),  # Image Orientation
  ]

  if dtype_changed:
    series_tag_values = series_tag_values + [
      ('0028|1053', str(slope)),  # rescale slope
      ('0028|1052', str(intercept)),  # rescale intercept
      ('0028|0100', '16'),  # bits allocated
      ('0028|0101', '16'),  # bits stored
      ('0028|0102', '15'),  # high bit
      ('0028|0103', '1')]  # pixel representation

  # Write slices to output directory
  list(map(lambda i: write_slices(writer, series_tag_values, new_img, outpath, i), range(new_img.GetDepth())))

def load_and_update_meta(inPath, outPath, tags):
  series_IDs = sitk.ImageSeriesReader.GetGDCMSeriesIDs(inPath)
  if not series_IDs:
    print("ERROR: given directory \""+inPpath+"\" does not contain a DICOM series.")
    return
    
  series_file_names = sitk.ImageSeriesReader.GetGDCMSeriesFileNames(inPath, series_IDs[0])

  series_reader = sitk.ImageSeriesReader()
  series_reader.SetFileNames(series_file_names)

  series_reader.MetaDataDictionaryArrayUpdateOn()
  series_reader.LoadPrivateTagsOn()
  image = series_reader.Execute()

  writer = sitk.ImageFileWriter()
  writer.KeepOriginalImageUIDOn()

  direction = image.GetDirection()

  modification_time = time.strftime("%H%M%S")
  modification_date = time.strftime("%Y%m%d")
  series_tag_values = [(k.replace("-", "|"), tags[k]) for k in tags.keys()]
  series_tag_values += [
    ("0008|0031", modification_time),  # Series Time
    ("0008|0021", modification_date),  # Series Date
    ('0020|0011', modification_date + modification_time), # Series Number
    ("0008|0008", "DERIVED\\SECONDARY"),  # Image Type
    ("0020|000e", "1.2.826.0.1.3680043.2.1125."
      + modification_date + ".1" + modification_time),  # Series Instance UID
    ("0020|0037", '\\'.join(map(str, (direction[0], direction[3], direction[6],
                                      direction[1], direction[4],
                                      direction[7])))),  # Image Orientation
  ]

  list(map(lambda i: write_slices(writer, series_tag_values, image, outPath, i), range(image.GetDepth())))

def upload_file(path, headers, authOrthanc, URL):

  f = open(path, "rb")
  content = f.read()
  f.close()

  try:
    resp = requests.post(URL + "/instances", data=content, auth=authOrthanc, headers=headers)
    if resp.status_code == 200:
      return resp.json()["ParentSeries"]
    else:
      return None

  except Exception as e:
    print(f"Error: {e}")
    return False

def import_dicom_to_orthanc(dcmpath, orthanc_url, username, password):
  success_count = 0
  total_file_count = 0

  headers = { 'content-type' : 'application/dicom' }
  authOrthanc = (username, password)

  URL = orthanc_url

  seriesId = None

  if os.path.isfile(dcmpath):
    # Upload a single file
    total_file_count += 1
    seriesId = upload_file(dcmpath, headers, URL)
    if not seriesId is None:
      success_count += 1

  else:
    # Recursively upload a directory
    for root, dirs, files in os.walk(dcmpath):
      for f in files:
        total_file_count += 1
        seriesId = upload_file(os.path.join(root, f), headers, authOrthanc, URL)
        if not seriesId is None:
          success_count += 1

  if success_count == total_file_count:
    print("\nSummary: all %d DICOM file(s) have been imported successfully" % success_count)
  else:
    print("\nSummary: %d out of %d files have been imported successfully as DICOM instances" % (success_count, total_file_count))

  return seriesId


import os, json
from .anonymize import anonymize

DATA = os.getenv('DATA', '/data')
MODULES = os.getenv('MODULES', '/modules')


def worker(jobName, headers, params, added_params, **kwargs):
  if "seriesIds" in headers:
    seriesIds = headers.get("seriesIds")
    for seriesId in seriesIds:
      dcmpath = os.path.join(DATA, headers.get("dcmpath"), seriesId)
      anonymize(dcmpath, params=params)
    



import os

def dicomAnonymizer(jobName, headers, params, added_params, **kwargs):
  injected_params = {}

  if "id" in headers:
    patientId = headers.get("id")
    injected_params["deleted"] = [os.path.join("dicom", patientId)]
  
  return True, injected_params
  
def systemcleaner(jobName, headers, params, added_params, **kwargs):
  return True, {}

def storageManager(jobName, headers, params, added_params, **kwargs):
  return True, {}



import os, requests, base64, json, time
from .lib import nifti_to_dicom, import_dicom_to_orthanc, load_and_update_meta
from datetime import datetime

DATA = os.getenv('DATA', '/data')
ORTHANC_REST_USERNAME = os.getenv('ORTHANC_REST_USERNAME', "anduin")
ORTHANC_REST_PASSWORD = os.getenv('ORTHANC_REST_PASSWORD', "anduin")
ORTHANC_REST_URL = os.getenv('ORTHANC_REST_URL', "http://orthanc:8042")
DEFAULT_STORE = os.getenv('ORTHANC_DEFUALT_STORE', "pac")
FORMATS = {
  "pdf": "data:application/pdf;base64,"
}

COPY_TAGS = [ '0002-0002', '0008-0016', '0008-0022', '0008-0023', '0008-0032', '0008-0033',
              '0008-0060', '0020-000d', '0020-0012', '0010-0010', '0010-0020', '0010-0030',
              '0010-0032', '0010-0040', '0020-0010', '0018-5100', '0018-0060', '0008-0090',
              '0008-0050', '0008-0020', '0008-0030', '0018-0015', '0040-0254', '0018-1030']

CLEAN_ORTHANC = os.getenv('CLEAN_ORTHANC', 0)

def load_header_codes():
  data = {}
  with open(os.path.join(os.path.dirname(os.path.realpath(__file__)),'header_codes.json')) as f:
    data = json.load(f)
  return data


def post_file_to_orthanc(filedata, filefmt, destination, added_tags, seriesId, action="store-data"):
  header = {'content-type': 'application/json'}
  authOrthanc = (ORTHANC_REST_USERNAME, ORTHANC_REST_PASSWORD)
  url = ORTHANC_REST_URL
  
  GET_studyinfo = url + "/series/" + seriesId + "/study"
  studyinfoResponse = requests.get(GET_studyinfo, auth=authOrthanc, headers=header)

  data = studyinfoResponse.json()

  tags = {
    "0405,0010" : "DCM-PROCESSOR",
    "0405,1001" : action,
    "0405,1003" : filefmt,
    "0405,1005" : "dcm-processor",
    "0405,1007" : destination,
    "0405,1009" : seriesId
  }

  tags.update(added_tags)

  payload = {
    "PrivateCreator": "DCM-PROCESSOR",
    "Parent": data["ID"],
    "Tags" : tags,
    "Content" : FORMATS[filefmt] + filedata
  }

  POST_pdf = url + "/tools/create-dicom"
  uploadpdfResponse = requests.post(POST_pdf, json=payload, auth=authOrthanc, headers=header)

  if uploadpdfResponse.status_code == 200:
    print("file successfully sent to orthanc")

    if int(CLEAN_ORTHANC) != 0:
      pdfData = uploadpdfResponse.json()
      newSeriesId = pdfData.get("ParentSeries")
      if not newSeriesId is None:
        DELETE_pdf = f"{url}/series/{newSeriesId}"
        requests.delete(DELETE_pdf, json=payload, auth=authOrthanc, headers=header)

  else:
    print("unable to sent file to orthanc")


def post_nifti_to_orthanc(path, filefmt, destination, added_tags, seriesId, base_folder, action="store-data"):
  header = {'content-type': 'application/json'}
  authOrthanc = (ORTHANC_REST_USERNAME, ORTHANC_REST_PASSWORD)
  url = ORTHANC_REST_URL
  
  GET_seriesinfo = url + "/series/" + seriesId
  seriesinfoResponse = requests.get(GET_seriesinfo, auth=authOrthanc, headers=header)
  data = seriesinfoResponse.json()

  outpath = os.path.join(base_folder, seriesId)

  os.system(f"mkdir -p {outpath}")
  
  if "Instances" in data:
    instances = data["Instances"]
    if len(instances) > 0:
      instanceId = str(instances[0])
      dicom_tags = requests.get(url + "/instances/" + instanceId + "/content", auth=authOrthanc, headers=header)
      tags = dicom_tags.json()
      dicom_meta = []
      
      for tag in tags:
        if tag not in COPY_TAGS:
          continue
        tag_url = f"{url}/instances/{instanceId}/content/{tag}"
        meta = requests.get(tag_url, auth=authOrthanc, headers=header)
        dicom_meta.append((str(tag).replace("-", "|"), meta.text))
        
      #Set Custom header
      dicom_meta += [("0405|0010", "DCM-PROCESSOR"), ("0405|1001", action), ("0405|1003", filefmt), ("0405|1005", "dcm-processor"), ("0405|1007", destination), ("0405|1009", seriesId)]

      #Set Series Description
      dicom_meta.append(("0008|103e", "Processed Information"))

      h_codes = load_header_codes()
      for k in added_tags:
        if k in h_codes:
          dicom_meta.append((h_codes[k], added_tags[k]))

      nifti_to_dicom(path, outpath ,dicom_meta)

      series_id = import_dicom_to_orthanc(outpath, url, ORTHANC_REST_USERNAME, ORTHANC_REST_PASSWORD)
      if (not series_id is None) and (not destination is None):
        time.sleep(2)
        patch_series_private_meta(series_id, {"0405-0010": "DCM-PROCESSOR", "0405-1001": action, "0405-1003": filefmt, "0405-1005": "dcm-processor", "0405-1007": destination, "0405-1009": seriesId})
        
  os.system(f"rm -rf {outpath}")


def post_dicom_to_orthanc(path, filefmt, destination ,added_tags, seriesId, base_folder, action="store-data"):
  header = {'content-type': 'application/json'}
  authOrthanc = (ORTHANC_REST_USERNAME, ORTHANC_REST_PASSWORD)
  url = ORTHANC_REST_URL
  
  GET_seriesinfo = url + "/series/" + seriesId
  seriesinfoResponse = requests.get(GET_seriesinfo, auth=authOrthanc, headers=header)
  data = seriesinfoResponse.json()

  if "Instances" in data:
    instances = data["Instances"]
    if len(instances) > 0:
      instanceId = str(instances[0])
      dicom_tags = requests.get(url + "/instances/" + instanceId + "/content", auth=authOrthanc, headers=header)
      tags = dicom_tags.json()
      dicom_meta = {}
      for tag in tags:
        if tag not in COPY_TAGS:
          continue

        tag_url = f"{url}/instances/{instanceId}/content/{tag}"
        meta = requests.get(tag_url, auth=authOrthanc, headers=header)
        dicom_meta[str(tag)] = meta.text

      h_codes = load_header_codes()
      for k in added_tags:
        if k in h_codes:
          dicom_meta[h_codes[k]] = added_tags[k]


      dicom_meta.update({"0405-0010": "DCM-PROCESSOR", "0405-1001": action, "0405-1003": filefmt, "0405-1005": "dcm-processor", "0405-1007": destination, "0405-1009": seriesId})

      output = os.path.join(base_folder, seriesId)
      os.system(f"mkdir -p {output}")

      load_and_update_meta(path, output, dicom_meta)
      series_id = import_dicom_to_orthanc(output, url, ORTHANC_REST_USERNAME, ORTHANC_REST_PASSWORD)

      os.system(f"rm -rf {output}")

      if (not series_id is None) and (not destination is None):
        time.sleep(2)
        patch_series_private_meta(series_id, {"0405-0010": "DCM-PROCESSOR", "0405-1001": action, "0405-1003": filefmt, "0405-1005": "dcm-processor", "0405-1007": destination, "0405-1009": seriesId})


def patch_series_private_meta(seriesId, tags):
  header = {'content-type': 'application/json'}
  authOrthanc = (ORTHANC_REST_USERNAME, ORTHANC_REST_PASSWORD)
  url = ORTHANC_REST_URL

  payloadModify = {"Replace" : tags, "PrivateCreator": "DCM-PROCESSOR"}
  POST_modify = url + "/series/" + seriesId + "/modify"

  resp = requests.post(POST_modify, json=payloadModify, auth=authOrthanc, headers=header)

  if resp.status_code == 200:
    newSeriesId = resp.json().get("ID")
    print("Successfully updated series", flush=True)
    DELETE_series = url + "/series/" + seriesId
    requests.delete(DELETE_series, auth=authOrthanc, headers=header)

    if (int(CLEAN_ORTHANC) != 0) and (not newSeriesId is None):
      time.sleep(2)
      DELETE_series = url + "/series/" + newSeriesId
      requests.delete(DELETE_series, auth=authOrthanc, headers=header)

  else:
    print("unable to update series tags", resp.status_code, flush=True)


def process_storage(storages, headers, params, added_params, **kwargs):
  now = datetime.now()
  current_time = now.strftime("%Y_%m_%d_%H_%M_%S")
  base_folder = os.path.join(DATA, f"storage_tmp_{current_time}")
  os.system(f"mkdir -p {base_folder}")

  for store in storages:
    if not isinstance(store, dict):
      continue

    if ("type" in store) and ("path" in store):
      f_type = store.get("type")
      f_type = str(f_type).lower()
      fullpath = os.path.join(DATA, store.get("path"))
      tags = {}

      destination = store.get("destination", DEFAULT_STORE)
      action = "store-data" if store.get("permanent", False) else "orthanc-only"

      if "tags" in store:
        t = store.get("tags")
        if isinstance(t, dict):
          tags = t

      if f_type in FORMATS:
        if not os.path.isfile(fullpath):
            continue
        try:
          fileobject = open(fullpath, "rb")
          filedata = str(base64.b64encode(fileobject.read()), 'utf-8')
          ref_series_id = store.get("seriesId")
          if ref_series_id is None:
            sids = headers.get("seriesIds", [])
            if len(sids) > 0:
              ref_series_id = sids[0]

          if not ref_series_id is None:
            post_file_to_orthanc(filedata, f_type, destination, tags, ref_series_id, action=action)
        except Exception as err:
          print(f"Error: {err}")

      elif f_type == "nifti":
        if not os.path.isfile(fullpath):
            continue

        ref_series_id = store.get("seriesId")
        if ref_series_id is None:
          sids = headers.get("seriesIds", [])
          if len(sids) > 0:
            ref_series_id = sids[0]
        
        if not ref_series_id is None:
          post_nifti_to_orthanc(fullpath, f_type, destination, tags, ref_series_id, base_folder, action=action)

      elif f_type == "dicom":
        if not os.path.isdir(fullpath):
            continue

        ref_series_id = store.get("seriesId")
        if ref_series_id is None:
          sids = headers.get("seriesIds", [])
          if len(sids) > 0:
            ref_series_id = sids[0]
        
        if not ref_series_id is None:
          post_dicom_to_orthanc(fullpath, f_type, destination, tags, ref_series_id, base_folder, action=action)
  
  os.system(f"rm -rf {base_folder}")


def worker(jobName, headers, params, added_params, **kwargs):
  try:
    for j in list(added_params.values()):
      if "storage" in j:
        tmp = j["storage"]
        storages = []

        if isinstance(tmp, list) or isinstance(tmp, tuple):
          storages = tmp
        elif isinstance(tmp, dict):
          storages = [tmp]

        process_storage(storages, headers, params, added_params, **kwargs)
  except:
    pass




#!/usr/bin/env python3

from __future__ import print_function
import os, argparse, json, dotenv
import inquirer as inq
import tempfile, git
from ruamel.yaml import YAML
from . import files as FILES
from . import worker as WORKER

INSTALLERS = ['local', 'git']
DEFUALT_DOCKER_SERVICES = ['worker', 'orthanc', 'mongo', 'dashboard', 'scheduler']

def handle_installer(inst, question, baseDir, isWorker=False):
  questions = []

  if inst == "git":
    questions.append(inq.Text("object_path", message=question + " (git url)"))

    if isWorker:
      questions.append(inq.Text("sub_path", message="Repository sub-folder", default="worker"))
    else:
      questions.append(inq.Text("sub_path", message="Repository sub-folder", default="service"))

  else:
    questions.append(inq.Path("object_path", message=question, exists=True, path_type=inq.Path.DIRECTORY, normalize_to_absolute_path=True))

  answers = inq.prompt(questions)

  if inst == "git":
    url = answers["object_path"]
    repo = git.Repo.clone_from(url, baseDir)
    repo.submodule_update()

    return os.path.join(baseDir, answers["sub_path"])

  return answers["object_path"]

def parse_args():
  parent_parser = argparse.ArgumentParser(description='A Command line tool for the dicom processor library', add_help=False)
  parent_parser.add_argument('action', metavar='action', type=str, help='action to be performed, options [create, init, install, remove, backup, build]')
  parent_parser.add_argument('object', metavar='object', type=str, help='action object, options [service, app]')
  args = parent_parser.parse_args()
  return args

def main():
  args = parse_args()
  action = args.action

  action_lower = str(action).lower()

  if action_lower == "create":
    handle_create(args)
  elif action_lower == "init":
    handle_init(args)
  elif action_lower == "install":
    handle_install(args)
  elif action_lower == "remove":
    handle_remove(args)
  elif action_lower == "backup":
    handle_backup(args)
  elif action_lower == "set":
    handle_set(args)
  elif action_lower == "start":
    handle_start(args)
  elif action_lower == "stop":
    handle_stop(args)
  elif action_lower == "restart":
    handle_restart(args)
  elif action_lower == "list":
    handle_list(args)
  else:
    print("Command not implemented!")

def write_json(data, filename):
  with open(filename, "w") as fp:
    json.dump(data, fp, indent=2)

def read_json(filename):
  with open(filename, "r") as fp:
    return json.load(fp)

  return None

def create_default_config():
  home_dir = os.path.expanduser("~")
  config_path = os.path.join(home_dir, ".dcm-processor", "config.json")
  os.system(f"mkdir -p {os.path.join(home_dir, '.dcm-processor')}")
  config = {
    "apps": {}
  }
  write_json(config, config_path)

def load_config(key=None):
  home_dir = os.path.expanduser("~")
  config_path = os.path.join(home_dir, ".dcm-processor", "config.json")
  config = {}

  if os.path.isfile(config_path):
    config = read_json(config_path)
  else:
    create_default_config()
    config = read_json(config_path)

  data = config

  if not key is None:
    if isinstance(key, str):
      ks = key.split(".")
      for k in ks:
        if not data is None:
          data = data.get(k)

  return data

def update_config(key, data):
  home_dir = os.path.expanduser("~")
  config_path = os.path.join(home_dir, ".dcm-processor", "config.json")
  config = {}

  if os.path.isfile(config_path):
    config = read_json(config_path)
  else:
    create_default_config()
    config = read_json(config_path)

  config[key] = data

  write_json(config, config_path)



def handle_create(args):
  object_lower = str(args.object).lower()
  if object_lower == "service":
    create_service(args)
  if object_lower == "app":
    create_app(args)
  if object_lower == "worker":
    create_worker(args)

def create_app(args):
  questions = [
    inq.Text("app_name", message="Enter app name"),
    inq.Path("app_path", message="Enter app base folder", exists=True, path_type=inq.Path.DIRECTORY, normalize_to_absolute_path=True),
    inq.Path("mapped_path", message="Enter app mapped folder base", exists=False, path_type=inq.Path.DIRECTORY, normalize_to_absolute_path=True),
    inq.Text("dicom_port", message="Enter orthanc DICOM port", default="4242"),
    inq.Text("browser_port", message="Enter orthanc browser port", default="8080"),
    inq.Text("dashboard_port", message="Enter jobs dashboard port", default="5000"),
    inq.Text("modality", message="Supported Modality (comma separated)", default="CT,MR"),
  ]
  
  answers = inq.prompt(questions)
  app_path = answers['app_path']
  mapped_path = answers['mapped_path']
  app_name = answers['app_name']
  dicom_port = answers['dicom_port']
  browser_port = answers['browser_port']
  dashboard_port = answers['dashboard_port']

  modality = answers['modality']

  existing_apps = load_config("apps")

  if app_name in existing_apps:
    print(f"There exist a configured app with name '{app_name}'")

  base_path = os.path.abspath(os.path.join(app_path, app_name))
  code_base_path = os.path.join(base_path)

  __initialize_app(base_path=base_path, mapped_folder=os.path.abspath(mapped_path))

  dotenv.set_key(os.path.join(code_base_path, ".env"), "BASEDIR", os.path.abspath(mapped_path))
  dotenv.set_key(os.path.join(code_base_path, ".env"), "ORTHANC_DICOM_PORT", dicom_port)
  dotenv.set_key(os.path.join(code_base_path, ".env"), "ORTHANC_BROWSER_PORT", browser_port)
  dotenv.set_key(os.path.join(code_base_path, ".env"), "SUPPORTED_MODALITY", modality)
  dotenv.set_key(os.path.join(code_base_path, ".env"), "DASHBOARD_PORT", dashboard_port)

  app_config = {
    "name": app_name,
    "base_dir": base_path,
    "mapped_path": os.path.abspath(mapped_path)
  }

  existing_apps[app_name] = app_config
  update_config("apps", existing_apps)
  init_app(args, app_name)

def __initialize_app(base_path, mapped_folder):
  code_base_path = os.path.join(base_path)

  os.system(f"mkdir -p {os.path.join(base_path)}")

  initial_files = {
    ".env": FILES.ENV,
    "docker-compose.yml": FILES.DOCKER_COMPOSE,
    "orthanc.json": FILES.ORTHANC,
    "build_volumes.sh": FILES.BUILD_VOLUMES,
    "build.sh": FILES.BUILD,
    "clean.sh": FILES.CLEAN,
    "init.sh": FILES.INIT,
    "rebuild.sh": FILES.REBUILD,
    "run.sh": FILES.RUN,
    "service.sh": FILES.SERVICE,
    "stop.sh": FILES.STOP,
    "settings.json": FILES.SETTINGS
  }

  ## Write initial files.
  for filename in initial_files:
    with open(os.path.join(code_base_path, filename), "w") as file:
      file.write(initial_files[filename])

  # Create and write dashboard config
  dashboard_dir = dotenv.get_key(os.path.join(code_base_path, ".env"), "DASHBOARD")
  if dashboard_dir is None:
    dashboard_dir = "dashboard"

  if str(dashboard_dir).startswith("/") or str(dashboard_dir).startswith("\\"):
    dashboard_dir = dashboard_dir[1:]
    
  dashboard_dir_full_path = os.path.join(mapped_folder, dashboard_dir)
  os.system(f"mkdir -p {dashboard_dir_full_path}")
  with open(os.path.join(dashboard_dir_full_path, "auth.json"), "w") as file:
    file.write(FILES.DASHBOARD_AUTH)

def create_worker(args):
  questions = [
    inq.Text("worker_name", message="Enter worker name"),
    inq.Path("worker_path", message="Enter workers folder", exists=True, path_type=inq.Path.DIRECTORY, normalize_to_absolute_path=True),
    inq.Text("worker_base_image", message="Enter worker base image", default="ubuntu:18.04"),
    inq.Text("worker_python_version", message="Enter worker python version", default="3.7"),
    inq.Text("channels", message="Enter job channels (comma separated)", default="default"),
    inq.Text("worker_description", message="Enter worker description"),
  ]

  answers = inq.prompt(questions)
  worker_name = answers['worker_name']
  python_version = answers['worker_python_version']
  base_image = answers["worker_base_image"]
  channels = answers["channels"]

  worker_settings = {
    "name": worker_name,
    "scripts": ["script.sh"],
    "entryScriptPaths": ["/scripts"],
    "baseImage": base_image,
    "environments": [
      {
        "name": "base",
        "requirements": ["requirements.txt"],
        "entryRequirementPaths": ["/requirements"],
        "channels": [s.strip() for s in str(channels).strip().split(",")],
        "pythonVersion": python_version
      }
    ]
  }

  fullpath = os.path.join(answers['worker_path'], answers['worker_name'])
  os.system(f"mkdir -p {fullpath}")
  
  os.system(f"touch {os.path.join(fullpath, 'requirements.txt')}")

  files = {
    "settings.json": json.dumps(worker_settings, indent=2),
    "script.sh": "#!/bin/bash"
  }

  for fn in files:
    with open(os.path.join(fullpath, fn), 'w') as txt:
      txt.write(files[fn])

def create_service(args):
  questions = [
    inq.Text("service_name", message="Enter service name"),
    inq.Path("service_path", message="Enter service folder", exists=True, path_type=inq.Path.DIRECTORY, normalize_to_absolute_path=True),
    inq.Text("service_description", message="Enter service description"),
  ]

  answers = inq.prompt(questions)
  service_name = answers['service_name']

  fullpath = os.path.join(answers['service_path'], answers['service_name'])
  modulePath = os.path.join(fullpath, "module")
  registryPath = os.path.join(fullpath, "registry")

  os.system(f"mkdir -p {modulePath}")
  os.system(f"mkdir -p {registryPath}")

  os.system(f"echo 'from .main import callback' > {os.path.join(registryPath, '__init__.py')}")

  os.system(f"echo 'from .main import worker' > {os.path.join(modulePath, '__init__.py')}")
  os.system(f"touch {os.path.join(modulePath, 'requirements.txt')}")
  os.system(f"touch {os.path.join(modulePath, 'script.sh')}")

  callback_code = [
    '\n',
    'def callback(jobName, headers, params, added_params, **kwargs):\n',
    '\tinjected_params = {"custom-data": "Some new data"}\n',
    '\n',
    '\t# Check header information etc. to see if you have to execute job\n'
    '\t# If Yes return True with the additional injected params\n'
    '\t# If No return False with additional injected params\n'
    '\n'
    '\treturn True, injected_params'
  ]
  worker_code = [
    'import os\n',
    '\n',
    'DATA = os.getenv("DATA")\n',
    'LOGS = os.getenv("LOGS")\n',
    'MODULES = os.getenv("MODULE")\n',
    '\n',
    'def worker(jobName, headers, params, added_params, **kwargs):\n',
    '\tprint(f"{jobName} can be handled here")\n',
    '\tprint(f"I can log info into the {LOGS} folder")\n',
    '\tprint(f"I can write and read from the {DATA} folder")\n',
    '\tprint(f"I can access other service modules from the {MODULES} folder")\n'
  ]
  settings = [
    {
      "jobName": service_name,
      "worker": f"{service_name}.worker",
      "callback": f"{service_name}.callback",
      "dependsOn": None,
      "channel": "default",
      "timeout": "1h",
      "params": {},
      "sortPosition": 0,
      "description": answers['service_description']
    }
  ]

  with open(os.path.join(registryPath, 'main.py'), 'w') as txt:
    txt.writelines(callback_code)

  with open(os.path.join(registryPath, 'settings.json'), 'w') as txt:
    json.dump(settings, txt, indent=2)

  with open(os.path.join(modulePath, 'main.py'), 'w') as txt:
    txt.writelines(worker_code)



def handle_init(args):
  object_lower = str(args.object).lower()
  if object_lower == "app":
    init_app(args)

def init_app(args, app_name=None):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return
  
  questions = [
    inq.List("pull_images", message="Pull docker images ?", choices=["Yes", "No"]),
  ]

  if app_name is None:
    questions = [
      inq.List("app_name", message="Select app to initial", choices=list(existing_apps.keys())),
    ] + questions

    answers = inq.prompt(questions)
    app_name = answers['app_name']
    pull_images = answers['pull_images'] == 'Yes'
  else:
    answers = inq.prompt(questions)
    pull_images = answers['pull_images'] == 'Yes'


  config = existing_apps.get(app_name)

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  build_script = os.path.join(base_dir, "build.sh")
  build_volumes_script = os.path.join(base_dir, "build_volumes.sh")
  run_script = os.path.join(base_dir, "run.sh")
  init_script = os.path.join(base_dir, "init.sh")
  stop_script = os.path.join(base_dir, "stop.sh")

  if not os.path.isfile(build_script):
    print("Missing build script!")
    return

  if not os.path.isfile(build_volumes_script):
    print("Missing build_volumes script!")
    return

  if not os.path.isfile(run_script):
    print("Missing run script!")
    return

  if not os.path.isfile(init_script):
    print("Missing init script!")
    return

  if not os.path.isfile(stop_script):
    print("Missing stop script!")
    return

  
  proxies = load_config("proxies")
  if not proxies is None:
    # Patch build script
    code_lines = ["#!/bin/bash\n","\n", "docker-compose build"] + [f" --build-arg {k}={proxies[k]}" for k in proxies.keys()] + [" --force-rm\n"] + ["docker-compose pull\n"]
    with open(build_script, "w") as sc:
      sc.writelines(code_lines)

  cwd = os.getcwd()

  cli_dir = os.path.dirname(os.path.realpath(__file__))
  cli_services = os.path.join(cli_dir, "services")
  print("copying base services....")
  os.system(f"cp -r {cli_services} {os.path.join(base_dir, 'services')}")

  os.chdir(os.path.join(base_dir))
  os.system(f"bash build_volumes.sh")

  if pull_images:
    os.system(f"bash build.sh")

  os.system(f"bash run.sh && bash stop.sh && bash init.sh")
  os.chdir(os.path.join(cwd))

  os.system(f"rm -rf {os.path.join(base_dir, 'services')}")



def handle_install(args):
  object_lower = str(args.object).lower()
  if object_lower == "service":
    install_service(args)
  if object_lower == "worker":
    install_worker(args)

def install_service(args):
  existing_apps = load_config("apps")
  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select app to install service to", choices=list(existing_apps.keys())),
    inq.Text("service_name", message="Enter service name"),
    inq.List("inst_type", message="Select installer", choices=INSTALLERS)
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']
  service_name = answers['service_name']
  inst_type = answers['inst_type']

  with tempfile.TemporaryDirectory() as baseDir:
    config = existing_apps.get(app_name)
    services = config.get("services", {})

    if service_name in services:
      print(f"A service with name {service_name} already exists!")
      return

    service_path = handle_installer(inst_type, "Enter service path", baseDir)

    if service_path is None:
      print("Unable install service!")
      return

    if config is None:
      print("Unable to load app configuration!")
      return

    base_dir = config.get("base_dir")
      
    if base_dir is None:
      print("Error in configuration filename")
      return

    service_script = os.path.join(base_dir, "service.sh")

    if not os.path.isfile(service_script):
      print("Missing build script!")
      return

    service_dir = os.path.abspath(service_path)

    cwd = os.getcwd()

    os.chdir(os.path.abspath(os.path.join(base_dir)))
    os.system(f"bash service.sh install {service_name} -p {service_dir}")

    if inst_type != "local":
      os.rmdir(service_dir)

    os.chdir(cwd)

    if not "services" in config:
      config["services"] = {}
    
    config["services"][service_name] = {
      "name": service_name,
      "base_dir": service_name
    }

    existing_apps[app_name] = config
    update_config("apps", existing_apps)

def install_worker(args):
  existing_apps = load_config("apps")
  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select app to install worker to", choices=list(existing_apps.keys())),
    inq.Text("worker_name", message="Enter worker name"),
    inq.List("inst_type", message="Select installer", choices=INSTALLERS),
    inq.List("start_worker", message="Start worker after installation ?", choices=["Yes", "No"], default="No"),
    inq.List("config_gpu", message="Configure GPU device reservation ?", choices=["Yes", "No"], default="No"),
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']
  worker_name = answers['worker_name']
  inst_type = answers['inst_type']
  config_gpu = answers['config_gpu']
  start_worker_after_install = answers['start_worker']
  device_config = {}

  with tempfile.TemporaryDirectory() as baseDir:
    worker_path = handle_installer(inst_type, "Enter worker path", baseDir, isWorker=True)

    if config_gpu == "Yes":
      device_answers = inq.prompt([
        inq.Text("device_cap", message="Enter device capabilities (comman separated list)", default="gpu"),
        inq.Text("driver", message="Enter device driver", default="nvidia"),
        inq.List("is_deviceid_or_count", message="How would you want to configure the reservation ?", choices=["Device Count", "Device ID"], default="Device Count"),
      ])

      caps = device_answers["device_cap"]
      driv = device_answers["driver"]
      is_dev_or_count = device_answers['is_deviceid_or_count']

      caps = str(caps).split(",")

      if len(caps) > 0:
        device_config["capabilities"] = caps

      if driv != "":
        device_config["driver"] = driv  

      if is_dev_or_count == "Device Count":
        count_answers = inq.prompt([
          inq.Text("count", message="Enter Device Count (integer) leave empty to use all devices"),
          inq.Text("device_options", message="Enter driver specific options (comma seperated key:value pairs)"),
        ])

        cnt = count_answers["count"]
        opts = count_answers["device_options"]

        if cnt != "":
          device_config["count"] = int(cnt)
        else:
          device_config["count"] = "all"
          
        if opts != "":
          opts = str(opts).split(",")         
          for opt in opts:
            keyval = opt.split(":")
            if len(keyval) > 1:

              if not "options" in device_config:
                device_config["options"] = {}
              
              device_config[keyval[0]] = keyval[1]

      else:
        deviceid_answers = inq.prompt([
          inq.Text("deviceids", message="Enter Device IDs (comman separated list)"),
          inq.Text("device_options", message="Enter driver specific options (comma seperated key:value pairs)"),
        ])

        devids = deviceid_answers["deviceids"]
        opts = deviceid_answers["device_options"]

        if devids != "":
          device_config["device_ids"] = str(devids).split(",")
          
        if opts != "":
          opts = str(opts).split(",")         
          for opt in opts:
            keyval = opt.split(":")
            if len(keyval) > 1:

              if not "options" in device_config:
                device_config["options"] = {}
              
              device_config[keyval[0]] = keyval[1]

    worker_name = str(worker_name).lower()

    if worker_name in DEFUALT_DOCKER_SERVICES:
      print(f"Worker name cannot be one of: {DEFUALT_DOCKER_SERVICES}")
      return

    config = existing_apps.get(app_name)

    if config is None:
      print("Unable to load app configuration!")
      return

    existing_workers = config.get("workers")

    if existing_workers is None:
      existing_workers = {}

    if worker_name in existing_workers:
      print(f"Worker with name {worker_name} already exist remove worker first!")
      return

    base_dir = config.get("base_dir")

    workers_dir = os.path.join(base_dir, "workers")

    os.system(f"mkdir -p {workers_dir}")

    # check required files
    settingsfile = os.path.join(worker_path, "settings.json")

    if not os.path.isfile(settingsfile):
      print("Worker missing settings.json!")
      return
    
    worker_settings = {}

    with open(settingsfile) as jsonFile:
      worker_settings = json.load(jsonFile)

    worker_base_dir = os.path.join(workers_dir, worker_name)

    __initialize_worker(base_path=worker_base_dir, settings=worker_settings)

    os.system(f"cp -r {os.path.join(worker_path)} {os.path.join(workers_dir, worker_name, 'settings')}")

    docker_compose_config = {
      'build': f"./workers/{worker_name}",
      'depends_on': ['mongo'],
      'environment': {
        'JOBS_CONNECTION': '${JOBS_CONNECTION}',
        'JOBS_DBNAME': '${JOBS_DBNAME}',
        'JOBS_COLNAME': '${JOBS_COLNAME}',
        'ORTHANC_REST_USERNAME': '${ORTHANC_REST_USERNAME}',
        'ORTHANC_REST_PASSWORD': '${ORTHANC_REST_PASSWORD}',
        'ORTHANC_REST_URL': '${ORTHANC_REST_URL}',
        'ORTHANC_DEFUALT_STORE': '${ORTHANC_DEFUALT_STORE}',
        'DATA': '/data',
        'MODULES': '/modules',
        'LOGS': '/logs'
      },
      'volumes': [
        '${BASEDIR}${MODULES}:/modules:cached',
        '${BASEDIR}${DATA}:/data:rw',
        '${BASEDIR}${LOGS}:/logs:rw'
      ]
    }

    if "image" in worker_settings:
      img = worker_settings.get("image")

      if not img is None:
        del docker_compose_config["build"]
        docker_compose_config["image"] = img

    if config_gpu == "Yes" and "capabilities" in device_config:
      docker_compose_config["deploy"] = {
        "resources": {
          "reservations": {
            "devices": [device_config]
          }
        }
      }

    yaml = YAML(typ="safe", pure=True)
    yaml.default_flow_style = False
    data = {}
    with open(os.path.join(base_dir, "docker-compose.yml"), 'r') as f: 
      data = yaml.load(f)
    
    with open(os.path.join(base_dir, "docker-compose.yml"), 'w') as f: 
      data["services"][str(worker_name)] = docker_compose_config
      yaml.dump(data, f)

    existing_workers[worker_name] = {
      "base_dir": worker_name,
    }

    config["workers"] = existing_workers
    existing_apps[app_name] = config
    update_config("apps", existing_apps)
    print(f"Worker successfully install as {worker_name}")

    cwd = os.getcwd()
    os.chdir(os.path.join(base_dir))

    if "build" in docker_compose_config:
      os.system(f"docker-compose build --force-rm {worker_name}")

    if start_worker_after_install == "Yes":
      os.system(f"docker-compose up -d {worker_name}")

    os.chdir(cwd)

def __initialize_worker(base_path, settings):
  code_base_path = os.path.join(base_path)

  os.system(f"mkdir -p {os.path.join(base_path)}")

  if "image" in settings:
    img = settings.get("image")
    if not img is None:
      return

  base_image = settings.get("baseImage", "ubuntu:18.04")

  initial_files = {
    "Dockerfile": WORKER.DOCKERFILE.replace("BASE_IMAGE", base_image),
    "entrypoint.sh": WORKER.ENTRYPOINT,
    "envs.py": WORKER.ENVS,
    "install.py": WORKER.INSTALL,
    "start.py": WORKER.START
  }

  ## Write initial files.
  for filename in initial_files:
    with open(os.path.join(code_base_path, filename), "w") as file:
      file.write(initial_files[filename])



def handle_remove(args):
  object_lower = str(args.object).lower()
  if object_lower == "service":
    remove_service(args)
  if object_lower == "app":
    remove_app(args)
  if object_lower == "worker":
    remove_worker(args)

def remove_app(args):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select app to remove", choices=list(existing_apps.keys())),
    inq.List("remove_base_dir", message="Remove app base directory ?", choices=["Yes", "No"]),
    inq.List("remove_mapped_dir", message="Remove app mapped folders directory ?", choices=["Yes", "No"]),
    inq.Text("backup_path", message="Enter backup folder (leave empty to skip backup)")
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']
  backup_path = answers['backup_path']
  remove_base = answers['remove_base_dir']
  remove_mapped = answers['remove_mapped_dir']

  config = existing_apps.get(app_name)

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")

  del existing_apps[app_name]
  update_config("apps", existing_apps)
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  clean_script = os.path.join(base_dir, "clean.sh")

  if not os.path.isfile(clean_script):
    print("Missing clean script!")
    return

  env_file = os.path.join(base_dir, ".env")
  if not os.path.isfile(env_file):
    print("Missing .env file!")
    return

  dotenv.load_dotenv(os.path.abspath(os.path.join(base_dir, '.env')))
  mapped_folder = os.environ.get("BASEDIR")

  if backup_path != "":
    backup_dir = os.path.abspath(backup_path)
    os.system(f"mkdir -p {backup_dir}")
    os.system(f"cp -r {os.path.abspath(os.path.join(base_dir))} {os.path.join(backup_dir, app_name)}")

    if not mapped_folder is None:
      os.system(f"cp -r {os.path.abspath(mapped_folder)} {os.path.join(backup_dir, app_name, 'mapped_folder')}")

  cwd = os.getcwd()

  os.chdir(os.path.abspath(os.path.join(base_dir)))
  os.system(f"bash clean.sh")
  os.chdir(cwd)

  if remove_base == "Yes":
    os.system(f"rm -rf {os.path.abspath(os.path.join(base_dir))}")
  
  if remove_mapped == "Yes":
    os.system(f"rm -rf {os.path.abspath(os.path.join(mapped_folder))}")

def remove_service(args):
  existing_apps = load_config("apps")
  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select app to remove service from", choices=list(existing_apps.keys())),
    inq.Text("service_name", message="Enter service name"),
    inq.Text("backup_path", message="Enter backup folder (leave empty to skip backup)")
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']
  service_name = answers['service_name']
  service_path = answers['backup_path']

  config = existing_apps.get(app_name)
  services = config.get("services", {})

  if not service_name in services:
    print(f"There is no service with name {service_name}!")
    return

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  service_script = os.path.join(base_dir, "service.sh")

  if not os.path.isfile(service_script):
    print("Missing service script!")
    return

  

  cwd = os.getcwd()
  os.chdir(os.path.abspath(os.path.join(base_dir)))

  if service_path != "":
    service_dir = os.path.abspath(service_path)
    os.system(f"bash service.sh remove {service_name} -b {service_dir}")
  else:
    os.system(f"bash service.sh remove {service_name}")

  os.chdir(cwd)

  del services[service_name]
  config["services"] = services
  existing_apps[app_name] = config
  update_config("apps", existing_apps)

def remove_worker(args):
  existing_apps = load_config("apps")
  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select app to remove worker from", choices=list(existing_apps.keys())),
    inq.Text("worker_name", message="Enter worker name")
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']
  worker_name = answers['worker_name']
  worker_name = str(worker_name).lower()

  config = existing_apps.get(app_name)

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  existing_workers = config.get("workers")
  if not worker_name in existing_workers:
    print(f"Worker with name {worker_name} does not exist!")
    return


  cwd = os.getcwd()
  os.chdir(os.path.join(base_dir))
  os.system(f"docker-compose rm -f -s {worker_name}")
  _, image_base = os.path.split(base_dir)
  os.system(f"docker rmi {image_base}_{worker_name}")
  os.system(f"rm -rf {os.path.join(base_dir, 'workers', worker_name)}")
  os.chdir(cwd)


  yaml = YAML(typ="safe", pure=True)
  yaml.default_flow_style = False
  data = {}
  with open(os.path.join(base_dir, "docker-compose.yml"), 'r') as f: 
    data = yaml.load(f)
  
  with open(os.path.join(base_dir, "docker-compose.yml"), 'w') as f: 
    del data["services"][worker_name]
    yaml.dump(data, f)

  del existing_workers[worker_name]
  config["workers"] = existing_workers
  existing_apps[app_name] = config
  update_config("apps", existing_apps)

  print(f"Worker successfully removed!")



def handle_backup(args):
  object_lower = str(args.object).lower()
  if object_lower == "service":
    backup_service(args)
  if object_lower == "app":
    backup_app(args)

def backup_service(args):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select app to backup service", choices=list(existing_apps.keys())),
    inq.Text("service_name", message="Enter service name"),
    inq.Text("backup_path", message="Enter backup folder")
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']
  service_name = answers['service_name']
  service_path = answers['backup_path']

  config = existing_apps.get(app_name)

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  service_script = os.path.join(base_dir, "service.sh")

  if not os.path.isfile(service_script):
    print("Missing service script!")
    return

  if service_path != "":
    service_dir = os.path.abspath(service_path)
    cwd = os.getcwd()
    os.chdir(os.path.abspath(os.path.join(base_dir)))
    os.system(f"bash service.sh backup {service_name} -b {service_dir}")
    os.chdir(cwd)
  else:
    print("Backup path not provided!")
    return

def backup_worker(args):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select app to backup worker", choices=list(existing_apps.keys())),
    inq.Text("worker_name", message="Enter worker name"),
    inq.Text("backup_path", message="Enter backup folder")
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']
  worker_name = answers['worker_name']
  backup_path = answers['backup_path']

  worker_name = str(worker_name).lower()

  config = existing_apps.get(app_name)

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  existing_workers = config.get("workers")

  if not worker_name in existing_workers:
    print(f"Worker with name {worker_name} does not exist!")
    return

  worker_dir = existing_workers[worker_name]["base_dir"]

  backup_path_full = os.path.abspath(os.path.join(backup_path, worker_name))

  os.system(f"cp -r {os.path.join(base_dir, 'workers', worker_dir)} {backup_path_full}")

  yaml = YAML(typ="safe", pure=True)
  yaml.default_flow_style = False
  data = {}

  with open(os.path.join(base_dir, "docker-compose.yml"), 'r') as f: 
    data = yaml.load(f)

  try:
    docker_config = data["services"][worker_name]
    write_json(docker_config, os.path.join(backup_path_full, "docker-compose.json"))
  except Exception as e:
    print(e)
    

  print("Worker successfully backed up!")

def backup_app(args):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select app to backup", choices=list(existing_apps.keys())),
    inq.Text("backup_path", message="Enter backup folder")
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']
  backup_path = answers['backup_path']

  config = existing_apps.get(app_name)

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  env_file = os.path.join(base_dir, ".env")
  if not os.path.isfile(env_file):
    print("Missing .env file!")
    return

  dotenv.load_dotenv(os.path.abspath(os.path.join(base_dir, '.env')))
  mapped_folder = os.environ.get("BASEDIR")

  if backup_path != "":
    backup_dir = os.path.abspath(backup_path)
    os.system(f"mkdir -p {backup_dir}")
    os.system(f"cp -r {os.path.abspath(os.path.join(base_dir))} {os.path.join(backup_dir, app_name)}")

    if not mapped_folder is None:
      os.system(f"cp -r {os.path.abspath(mapped_folder)} {os.path.join(backup_dir, app_name, 'mapped_folder')}")



def handle_start(args):
  object_lower = str(args.object).lower()
  if object_lower == "app":
    start_app(args)

def start_app(args, app_name=None):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  if app_name is None:
    questions = [
      inq.List("app_name", message="Select app to start", choices=list(existing_apps.keys())),
    ]

    answers = inq.prompt(questions)
    app_name = answers['app_name']

  config = existing_apps.get(app_name)

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  run_script = os.path.join(base_dir, "run.sh")

  if not os.path.isfile(run_script):
    print("Missing run script!")
    return

  cwd = os.getcwd()

  os.chdir(os.path.join(base_dir))
  os.system(f"bash run.sh")
  os.chdir(os.path.join(cwd))



def handle_stop(args):
  object_lower = str(args.object).lower()
  if object_lower == "app":
    stop_app(args)

def stop_app(args, app_name=None):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  if app_name is None:
    questions = [
      inq.List("app_name", message="Select app to start", choices=list(existing_apps.keys())),
    ]

    answers = inq.prompt(questions)
    app_name = answers['app_name']

  config = existing_apps.get(app_name)

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  stop_script = os.path.join(base_dir, "stop.sh")

  if not os.path.isfile(stop_script):
    print("Missing stop script!")
    return

  cwd = os.getcwd()

  os.chdir(os.path.join(base_dir))
  os.system(f"bash stop.sh")
  os.chdir(os.path.join(cwd))




def handle_restart(args):
  object_lower = str(args.object).lower()
  if object_lower == "app":
    restart_app(args)

def restart_app(args, app_name=None):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  if app_name is None:
    questions = [
      inq.List("app_name", message="Select app to start", choices=list(existing_apps.keys())),
    ]

    answers = inq.prompt(questions)
    app_name = answers['app_name']

  config = existing_apps.get(app_name)

  if config is None:
    print("Unable to load app configuration!")
    return

  base_dir = config.get("base_dir")
    
  if base_dir is None:
    print("Error in configuration filename")
    return

  run_script = os.path.join(base_dir, "run.sh")

  if not os.path.isfile(run_script):
    print("Missing run script!")
    return

  cwd = os.getcwd()

  os.chdir(os.path.join(base_dir))
  os.system(f"bash run.sh")
  os.chdir(os.path.join(cwd))




def handle_set(args):
  object_lower = str(args.object).lower()
  if object_lower == "proxy":
    set_proxy(args)
  
def set_proxy(args):
  questions = [
    inq.Text("http_proxy", message="Enter http proxy"),
    inq.Text("https_proxy", message="Enter https proxy"),
    inq.Text("ftp_proxy", message="Enter ftp proxy"),
    inq.Text("no_proxy", message="Enter no proxy urls (comma separated)"),
  ]

  answers = inq.prompt(questions)
  
  p_settings = {
    "http_proxy": answers["http_proxy"],
    "https_proxy": answers["https_proxy"],
    "ftp_proxy": answers["ftp_proxy"],
    "no_proxy": answers["no_proxy"],
  }

  update_config("proxies", p_settings)
  print("Proxy settings updated!")
  
def set_trusted_hosts(args):
  questions = [
    inq.Text("trusted_hosts", message="Enter pip trusted hosts"),
  ]

  answers = inq.prompt(questions)
  
  th = answers["trusted_hosts"]
  th = str(th).split(",")
  
  update_config("trusted_hosts", th)
  print("Trusted hosts updated!")


def handle_list(args):
  object_lower = str(args.object).lower()
  if object_lower == "app" or object_lower == "apps":
    list_apps(args)
  if object_lower == "service" or object_lower == "services":
    list_services(args)
  if object_lower == "worker" or object_lower == "workers":
    list_workers(args)

def list_apps(args):
  existing_apps = load_config("apps")
  print("List of existing applications")
  print("--------------------------------------------------------")
  for k in existing_apps:
    item = existing_apps[k]
    print(f"{'{0: <25}'.format(item.get('name'))}\t{item.get('base_dir')}")

def list_services(args):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select application", choices=list(existing_apps.keys())),
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']

  config = existing_apps.get(app_name)
  services = config.get("services", {})

  print(f"List of services for {app_name}")
  print("--------------------------------------------------------")

  mapped_path = config.get("mapped_path", "")

  for k in services:
    item = services[k]
    print(f"{'{0: <25}'.format(item.get('name'))}\t{os.path.join(mapped_path, 'modules', item.get('base_dir'))}\t{os.path.join(mapped_path, 'registry', item.get('base_dir'))}")

def list_workers(args):
  existing_apps = load_config("apps")

  if len(existing_apps.keys()) == 0:
    print("There are no existing apps")
    return

  questions = [
    inq.List("app_name", message="Select application", choices=list(existing_apps.keys())),
  ]

  answers = inq.prompt(questions)
  app_name = answers['app_name']

  config = existing_apps.get(app_name)
  workers = config.get("workers", {})

  print(f"List of workers for {app_name}")
  print("--------------------------------------------------------")

  base_dir = config.get("base_dir", "")

  for k in workers:
    item = workers[k]
    print(f"{'{0: <25}'.format(k)}\t{os.path.join(base_dir, 'workers', item.get('base_dir'))}")

if __name__ == "__main__":
  main()


import json, chardet

class BytesEncoder(json.JSONEncoder):
  def default(self, obj):
    if isinstance(obj, bytes):
      encoding = chardet.detect(obj).get('encoding')
      return obj.decode(encoding, errors='replace')
    return json.JSONEncoder.default(self, obj)
  

def remove_strange_tags(data):
  tags = ["PixelData"]
  for tag in tags:
    if tag in data:
      del data[tag]
      
  return data


import os, shutil, requests

DATA = os.getenv('DATA', '/data')
CLEAN_ORTHANC = os.getenv('CLEAN_ORTHANC', 0)

def worker(jobName, headers, params, added_params, **kwargs):
  try:
    for j in list(added_params.values()):
      if "deleted" in j:
        tmp = j["deleted"]
        fns = []

        if isinstance(tmp, list) or isinstance(tmp, tuple):
          fns = tmp
        elif isinstance(tmp, str):
          fns = [tmp]

        for fn in fns:
          try:
            ffn = os.path.join(DATA, fn)
            if os.path.exists(ffn):
              if os.path.isfile(ffn):
                os.remove(os.path.join(DATA, fn))
              elif os.path.isdir(ffn):
                shutil.rmtree(ffn)
          except:
            pass
  except:
    pass
  
  if int(CLEAN_ORTHANC) != 0:
    clean_orthanc(jobName, headers, params, added_params, **kwargs)

def clean_orthanc(jobName, headers, params, added_params, **kwargs):
  ORTHANC_REST_USERNAME = os.getenv('ORTHANC_REST_USERNAME', "anduin")
  ORTHANC_REST_PASSWORD = os.getenv('ORTHANC_REST_PASSWORD', "anduin")
  ORTHANC_REST_URL = os.getenv('ORTHANC_REST_URL', "http://orthanc:8042")

  header = {'content-type': 'application/json'}
  authOrthanc = (ORTHANC_REST_USERNAME, ORTHANC_REST_PASSWORD)
  url = ORTHANC_REST_URL

  seriesIds = headers.get("seriesIds", [])
  for seriesId in seriesIds:
    if not seriesId is None:
      resp = requests.get(f"{url}/series/{seriesId}", auth=authOrthanc, headers=header)
      if resp.status_code == 200:
        studyId = resp.json().get("ParentStudy")
        patientId = None
        resp = requests.get(f"{url}/studies/{studyId}", auth=authOrthanc, headers=header)
        if resp.status_code == 200:
          study = resp.json()
          series = study.get("Series", [])
          for sId in series:
            if str(sId) == str(seriesId):
              try:
                requests.delete(f"{url}/series/{sId}", auth=authOrthanc, headers=header)
              except:
                pass
            else:
              resp = requests.get(f"{url}/series/{sId}", auth=authOrthanc, headers=header)
              if resp.status_code == 200:
                s = resp.json()
                instances = s.get("Instances")
                if len(instances) > 0:
                  instanceId = instances[0]
                  resp = requests.get(f"{url}/instances/{instanceId}", auth=authOrthanc, headers=header)
                  if resp.status_code == 200:
                    inst = resp.json()
                    ref = inst.get("ReferenceSeries")
                    if (not ref is None) and str(ref) == str(seriesId):
                      try:
                        requests.delete(f"{url}/series/{sId}", auth=authOrthanc, headers=header)
                      except:
                        pass

          resp = requests.get(f"{url}/studies/{study.get('ID')}", auth=authOrthanc, headers=header)
          patientId = study.get("ParentPatient")
          if resp.status_code == 200:
            study = resp.json()
            series = study.get("Series")
            if len(series) == 0:
              try:
                requests.delete(f"{url}/studies/{study.get('ID')}", auth=authOrthanc, headers=header)
              except:
                pass

        if not patientId is None:
          resp = requests.get(f"{url}/patients/{patientId}", auth=authOrthanc, headers=header)
          if resp.status_code == 200:
            patient = resp.json()
            studies = patient.get('Studies')
            print("studies", studies, flush=True)

            if len(studies) == 0:
              try:
                requests.delete(f"{url}/patients/{patient.get('ID')}", auth=authOrthanc, headers=header)
              except:
                pass
        else:
          print("Undefined patient ID")


BUILD = """
#!/bin/bash

docker-compose build --force-rm
docker-compose pull
"""

DASHBOARD_AUTH = """
{ "dcm-processor": "dcm-processor" }
"""

SETTINGS = """
{
  "headerFields": ["SeriesInstanceUID", "ImagePositionPatient", "ImageOrientationPatient", "PerformedProcedureStepDescription", "seriesId", "dcmpath", "SeriesDescription", "ContrastBolusAgent", "Modality", "ImageType", "PatientID", "SeriesNumber", "StudyDate"],
  "preServices": [
    {
      "jobName": "dicomAnonymizer",
      "worker": "base.DicomAnonymizerService.worker",
      "callback": "base.dicomAnonymizer",
      "dependsOn": null,
      "channel": "default",
      "timeout": "1h",
      "params": {
        "fhir": {
          "interface_url": "",
          "auth_server_url": "",
          "client_id": "",
          "client_secret": "",
          "identifier_system": ""
        }
      },
      "sortPosition": 0,
      "description": "Dicom Anonymizer Service"
    }
  ],
  "postServices":[
    {
      "jobName": "storageManager",
      "worker": "base.storageManager.worker",
      "callback": "base.storageManager",
      "dependsOn": null,
      "channel": "default",
      "timeout": "1h",
      "params": {},
      "sortPosition": 0,
      "description": "Storage Service"
    },
    {
      "jobName": "systemClearner",
      "worker": "base.systemcleaner.worker",
      "callback": "base.systemcleaner",
      "dependsOn": "storageManager",
      "channel": "default",
      "timeout": "1h",
      "params": {},
      "sortPosition": 1,
      "description": "System Cleaning Service"
    }
  ]
}
"""

CLEAN = """
#!/bin/bash

docker-compose down

SCRIPT=$(readlink -f "$0")
BASEDIR=$(dirname "$SCRIPT")

for i in $(echo $BASEDIR | tr "/" "\n")
do
  BASE="${i}"
done

filter="${BASE}_*:*"
docker rmi -f $(docker images --filter=reference="$filter" -q)

docker volume rm "${BASE}_mongo-data"
"""

BUILD_VOLUMES = """
#!/bin/bash
set -o allexport; source .env; set +o allexport

[ -z "$BASEDIR" ] && echo "set BASEDIR variable in the .env file" && exit 1
[ -z "$MODULES" ] && echo "set MODULES variable in the .env file" && exit 1
[ -z "$REGISTRY" ] && echo "set REGISTRY variable in the .env file" && exit 1
[ -z "$DATA" ] && echo "set DATA variable in the .env file" && exit 1
[ -z "$LOGS" ] && echo "set  variable in the .env file" && exit 1

mkdir -p "$BASEDIR/$MODULES"
mkdir -p "$BASEDIR/$REGISTRY"
mkdir -p "$BASEDIR/$DATA"
mkdir -p "$BASEDIR/$LOGS"
"""

RUN = """
#!/bin/bash

docker-compose down

docker-compose up -d orthanc scheduler

sleep 5

if [ -z "$1" ]
then
  docker-compose up -d
else
  docker-compose up $1 -d
fi
"""

REBUILD = """
#!/bin/bash

docker-compose stop $1
docker-compose rm $1


SCRIPT=$(readlink -f "$0")
BASEDIR=$(dirname "$SCRIPT")

for i in $(echo $BASEDIR | tr "/" "\n")
do
  BASE="${i}"  
done

docker rmi "${BASE}_$1"

docker-compose build $1

docker-compose up -d $1
"""

STOP = """
#!/bin/bash

docker-compose down
"""

INIT = """
#!/bin/bash

bash service.sh install base -p ./services/base
bash service.sh install dcm2nii -p ./services/dcm2nii

"""

SERVICE = """
#!/bin/bash

set -o allexport; source .env; set +o allexport

compose="docker-compose -f docker-compose.yml"

usage()
{
  echo "Usage: $0 [-h] [action] [servicename] [ -p SERVICE_PATH ] [ -b BACKUP_PATH  ]"
  echo "actions: [install | remove | backup]"
  echo "-h      To show this help message"
  echo "-p      Path to the folder where the service data is. this is a directory with two sub-directories [registry and module]"
  echo "-b      Path to the folder where the service backup will be place. supported by [remove and backup] actions"
  exit 2
}

set_variable()
{
  local varname=$1
  shift
  if [ -z "${!varname}" ]; then
    eval "$varname=\"$@\""
  else
    echo "Error: $varname already set"
    usage
  fi
}

valid_action_type()
{
    case "$1" in
    "install"|"remove"|"backup")
        return 0;;
    *)
        echo "Action $1 is not supported"
        return 1;;
    esac
}

[ -z "$BASEDIR" ] && echo "set BASEDIR variable in the .env file" && exit 1
[ -z "$MODULES" ] && echo "set MODULES variable in the .env file" && exit 1
[ -z "$REGISTRY" ] && echo "set REGISTRY variable in the .env file" && exit 1

unset ACTION SERVICENAME SERVICEPATH BACKUPPATH

if ! valid_action_type "$1"; then
    usage
    exit 1
fi

ACTION=$1
SERVICENAME=$2
shift 2

[ -z "$SERVICENAME" ] && usage && exit 1

[[ $SERVICENAME == -* ]] && usage && exit 1


# Install
if [ $ACTION = "install" ]
then

  while getopts ':p:?h:' c; do
    case $c in
      p) set_variable SERVICEPATH $OPTARG ;;
      h|?) usage ;;
    esac
  done

  if [ -z "$SERVICEPATH" ]
  then
    read -p "Enter Service Path :" SERVICEPATH
  fi

  echo "stopping containers..."
  # bash stop.sh
  if [ -d "$BASEDIR/$MODULES/$SERVICENAME" ]
  then
    echo "removing existing service module entry..."
    rm -rf "$BASEDIR/$MODULES/$SERVICENAME"
  fi

  if [ -d "$BASEDIR/$REGISTRY/$SERVICENAME" ]
  then
    echo "removing existing service registry entry..."
    rm -rf "$BASEDIR/$REGISTRY/$SERVICENAME"
  fi

  echo "copying module folder..."
  cp -r "$SERVICEPATH/module" "$BASEDIR/$MODULES/$SERVICENAME"
  echo "copying registry folder..."
  cp -r "$SERVICEPATH/registry" "$BASEDIR/$REGISTRY/$SERVICENAME"
  echo "starting workers..."
  # bash run.sh
fi



# Backup
if [ $ACTION = "backup" ]
then

  while getopts ':b:?h:' c; do
    case $c in
      b) set_variable BACKUPPATH $OPTARG ;;
      h|?) usage ;;
    esac
  done

  if [ -z "$BACKUPPATH" ]
  then
    read -p "Enter Backup Path :" SERVICEPATH
  fi

  mkdir -p "$BACKUPPATH/$SERVICENAME"
  echo "Copying registry entry"
  cp -r "${BASEDIR}${REGISTRY}/$SERVICENAME" "$BACKUPPATH/$SERVICENAME/registry"
  echo "Copying modules entry"
  cp -r "${BASEDIR}${MODULES}/$SERVICENAME" "$BACKUPPATH/$SERVICENAME/module"
fi




# Remove
if [ $ACTION = "remove" ]
then

  while getopts ':b:?h:' c; do
    case $c in
      b) set_variable BACKUPPATH $OPTARG ;;
      h|?) usage ;;
    esac
  done

  if [ -z "$BACKUPPATH" ]
  then
    echo "removing registry entry"
    rm -rf "$BASEDIR/$REGISTRY/$SERVICENAME"
    echo "removing modules entry"
    rm -rf "$BASEDIR/$MODULES/$SERVICENAME"
  else
    mkdir -p "$BACKUPPATH/$SERVICENAME"
    echo "Moving registry entry"
    mv "$BASEDIR/$REGISTRY/$SERVICENAME" "$BACKUPPATH/$SERVICENAME/registry"
    echo "Moving modules entry"
    mv "$BASEDIR/$MODULES/$SERVICENAME" "$BACKUPPATH/$SERVICENAME/module"
  fi

fi
"""

ENV = """
BASEDIR=./mapped_folders
MODULES=/modules
DATA=/data
REGISTRY=/registry
LOGS=/logs
DASHBOARD=/dashboard

SCHEDULER_HOST=http://scheduler
SCHEDULER_PORT=5000
SCHEDULER_SETTINGS=/settings.json

ORTHANC_REST_USERNAME=admin
ORTHANC_REST_PASSWORD=admin
ORTHANC_REST_URL=http://orthanc:8042
ORTHANC_DEFUALT_STORE=pac
CLEAN_ORTHANC=0

ORTHANC_DICOM_PORT=4242
ORTHANC_BROWSER_PORT=8042
DASHBOARD_PORT=5000
SUPPORTED_MODALITY=CT,MR
ACCEPTED_FILES=primary
JUNK_FILES=sbi,surv,bersi,racker,ssde,results,mip,mono,spectal,scout,localizer,lokali,konturen,sectrareconstruction,zeffect,iodoinekein,smartplan,doseinf

JOBS_CONNECTION=mongodb://mongo:27017/
JOBS_DBNAME=jobs
JOBS_COLNAME=jobs
DEFUALT_CHANNEL=default

IMAGE_VERSION=0.0.2
"""

DOCKER_COMPOSE = """
version: '3.7'

volumes:
  mongo-data:
    driver: local

services:
  mongo:
    image: mongo
    volumes:
      - mongo-data:/data/db
    expose:
      - "27017"

  worker:
    image: giesekow/dcm-processor-worker:${IMAGE_VERSION}
    depends_on:
      - mongo
    environment:
      JOBS_CONNECTION: ${JOBS_CONNECTION}
      JOBS_DBNAME: ${JOBS_DBNAME}
      JOBS_COLNAME: ${JOBS_COLNAME}
      ORTHANC_REST_USERNAME: ${ORTHANC_REST_USERNAME}
      ORTHANC_REST_PASSWORD: ${ORTHANC_REST_PASSWORD}
      ORTHANC_REST_URL: ${ORTHANC_REST_URL}
      ORTHANC_DEFUALT_STORE: ${ORTHANC_DEFUALT_STORE}
      CLEAN_ORTHANC: ${CLEAN_ORTHANC}
      DATA: /data
      MODULES: /modules
      LOGS: /logs
    volumes:
      - ${BASEDIR}${MODULES}:/modules:cached
      - ${BASEDIR}${DATA}:/data:rw
      - ${BASEDIR}${LOGS}:/logs:rw

  scheduler:
    image: giesekow/dcm-processor-scheduler:${IMAGE_VERSION}
    depends_on:
      - mongo
    environment:
      JOBS_CONNECTION: ${JOBS_CONNECTION}
      JOBS_DBNAME: ${JOBS_DBNAME}
      JOBS_COLNAME: ${JOBS_COLNAME}
      DATA: /data
      REGISTRY: /registry
      LOGS: /logs
      DEFUALT_CHANNEL: ${DEFUALT_CHANNEL}
      SETTINGS: ${SCHEDULER_SETTINGS}
    volumes:
      - ${BASEDIR}${REGISTRY}:/registry:cached
      - ${BASEDIR}${DATA}:/data:rw
      - ${BASEDIR}${LOGS}:/logs:rw
      - ./settings.json:/settings.json
  
  dashboard:
    image: giesekow/mongo-qas-dashboard:latest
    expose:
      - "5000"
    ports:
      - ${DASHBOARD_PORT}:5000
    depends_on:
      - mongo
    environment:
      DB_CONNECTION: ${JOBS_CONNECTION}
      DB_NAME: ${JOBS_DBNAME}
      COL_NAME: ${JOBS_COLNAME}
    volumes:
      - ${BASEDIR}${DASHBOARD}:/data

  orthanc:
    image: giesekow/dcm-processor-orthanc:${IMAGE_VERSION}
    command: /run/secrets/  # Path to the configuration files (stored as secrets)
    expose:
      - "4242"
      - "8042"
    ports:
      - ${ORTHANC_DICOM_PORT}:4242
      - ${ORTHANC_BROWSER_PORT}:8042
    volumes:
      - ${BASEDIR}/orthanc/db:/var/lib/orthanc/db:rw                # Persitent Orthanc DB
      - ${BASEDIR}/orthanc/dblight:/var/lib/orthanc/dblight:rw      # Persitent Orthanc DB
      - ${BASEDIR}${DATA}:/tmp/nifti:rw                             # Temp folder
    secrets:
      - orthanc.json
    environment:
      SCHEDULER_HOST: ${SCHEDULER_HOST}
      SCHEDULER_PORT: ${SCHEDULER_PORT}
      SUPPORTED_MODALITY: ${SUPPORTED_MODALITY}
      JUNK_FILES: ${JUNK_FILES}
      ACCEPTED_FILES: ${ACCEPTED_FILES}

secrets:
  orthanc.json:
    file: ./orthanc.json
"""

ORTHANC = """
{
  /**
   * General configuration of Orthanc
   **/

  // The logical name of this instance of Orthanc. This one is
  // displayed in Orthanc Explorer and at the URI "/system".
  "Name": "dcm-processor",

  // Path to the directory that holds the heavyweight files (i.e. the
  // raw DICOM instances). Backslashes must be either escaped by
  // doubling them, or replaced by forward slashes "/".
  "StorageDirectory": "/var/lib/orthanc/db",

  // Path to the directory that holds the SQLite index (if unset, the
  // value of StorageDirectory is used). This index could be stored on
  // a RAM-drive or a SSD device for performance reasons.
  "IndexDirectory": "/var/lib/orthanc/dblight",

  // Path to the directory where Orthanc stores its large temporary
  // files. The content of this folder can be safely deleted if
  // Orthanc once stopped. The folder must exist. The corresponding
  // filesystem must be properly sized, given that for instance a ZIP
  // archive of DICOM images created by a job can weight several GBs,
  // and that there might be up to "min(JobsHistorySize,
  // MediaArchiveSize)" archives to be stored simultaneously. If not
  // set, Orthanc will use the default temporary folder of the
  // operating system (such as "/tmp/" on UNIX-like systems, or
  // "C:/Temp" on Microsoft Windows).
  "TemporaryDirectory": "/tmp",

  // Enable the transparent compression of the DICOM instances
  "StorageCompression": false,

  // Maximum size of the storage in MB (a value of "0" indicates no
  // limit on the storage size)
  "MaximumStorageSize": 0,

  // Maximum number of patients that can be stored at a given time
  // in the storage (a value of "0" indicates no limit on the number
  // of patients)
  "MaximumPatientCount": 0,

  // List of paths to the custom Lua scripts that are to be loaded
  // into this instance of Orthanc

  "PythonScript": "/scripts/Callback.py",
  "PythonVerbose": false,

  // List of paths to the plugins that are to be loaded into this
  // instance of Orthanc (e.g. "./libPluginTest.so" for Linux, or
  // "./PluginTest.dll" for Windows). These paths can refer to
  // folders, in which case they will be scanned non-recursively to
  // find shared libraries. Backslashes must be either escaped by
  // doubling them, or replaced by forward slashes "/".
  "Plugins": [
    "/usr/share/orthanc/plugins", "/usr/local/share/orthanc/plugins"
  ],

  // Maximum number of processing jobs that are simultaneously running
  // at any given time. A value of "0" indicates to use all the
  // available CPU logical cores. To emulate Orthanc <= 1.3.2, set
  // this value to "1".
  "ConcurrentJobs": 2,


  /**
   * Configuration of the HTTP server
   **/

  // Enable the HTTP server. If this parameter is set to "false",
  // Orthanc acts as a pure DICOM server. The REST API and Orthanc
  // Explorer will not be available.
  "HttpServerEnabled": true,

  // HTTP port for the REST services and for the GUI
  "HttpPort": 8042,

  // When the following option is "true", if an error is encountered
  // while calling the REST API, a JSON message describing the error
  // is put in the HTTP answer. This feature can be disabled if the
  // HTTP client does not properly handles such answers.
  "HttpDescribeErrors": true,

  // Enable HTTP compression to improve network bandwidth utilization,
  // at the expense of more computations on the server. Orthanc
  // supports the "gzip" and "deflate" HTTP encodings.
  "HttpCompressionEnabled": true,



  /**
   * Configuration of the DICOM server
   **/

  // Enable the DICOM server. If this parameter is set to "false",
  // Orthanc acts as a pure REST server. It will not be possible to
  // receive files or to do query/retrieve through the DICOM protocol.
  "DicomServerEnabled": true,

  // The DICOM Application Entity Title (cannot be longer than 16
  // characters)
  "DicomAet": "DCM-PROCESSOR",

  // Check whether the called AET corresponds to the AET of Orthanc
  // during an incoming DICOM SCU request
  "DicomCheckCalledAet": false,

  // The DICOM port
  "DicomPort": 4242,

  // The default encoding that is assumed for DICOM files without
  // "SpecificCharacterSet" DICOM tag, and that is used when answering
  // C-Find requests (including worklists). The allowed values are
  // "Ascii", "Utf8", "Latin1", "Latin2", "Latin3", "Latin4",
  // "Latin5", "Cyrillic", "Windows1251", "Arabic", "Greek", "Hebrew",
  // "Thai", "Japanese", "Chinese", "JapaneseKanji", "Korean", and
  // "SimplifiedChinese".
  "DefaultEncoding": "Latin1",

  // The transfer syntaxes that are accepted by Orthanc C-Store SCP
  "DeflatedTransferSyntaxAccepted": true,
  "JpegTransferSyntaxAccepted": true,
  "Jpeg2000TransferSyntaxAccepted": true,
  "JpegLosslessTransferSyntaxAccepted": true,
  "JpipTransferSyntaxAccepted": true,
  "Mpeg2TransferSyntaxAccepted": true,
  "RleTransferSyntaxAccepted": true,
  "Mpeg4TransferSyntaxAccepted": true, // New in Orthanc 1.6.0

  // Whether Orthanc accepts to act as C-Store SCP for unknown storage
  // SOP classes (aka. "promiscuous mode")
  "UnknownSopClassAccepted": false,

  // Set the timeout (in seconds) after which the DICOM associations
  // are closed by the Orthanc SCP (server) if no further DIMSE
  // command is received from the SCU (client).
  "DicomScpTimeout": 30,



  /**
   * Security-related options for the HTTP server
   **/

  // Whether remote hosts can connect to the HTTP server
  "RemoteAccessAllowed": true,

  // Whether or not SSL is enabled
  "SslEnabled": false,

  // Path to the SSL certificate in the PEM format (meaningful only if
  // SSL is enabled)
  "SslCertificate": "certificate.pem",

  // Whether or not the password protection is enabled (using HTTP
  // basic access authentication). Starting with Orthanc 1.5.8, if
  // "AuthenticationEnabled" is not explicitly set, authentication is
  // enabled iff. remote access is allowed (i.e. the default value of
  // "AuthenticationEnabled" equals that of "RemoteAccessAllowed").
  /**
     "AuthenticationEnabled" : true,
   **/

  // The list of the registered users. Because Orthanc uses HTTP
  // Basic Authentication, the passwords are stored as plain text.
  "RegisteredUsers": {
    "admin": "admin"
  },



  /**
   * Network topology
   **/

  // The list of the known DICOM modalities
  "DicomModalities": {
    /**
     * Uncommenting the following line would enable Orthanc to
     * connect to an instance of the "storescp" open-source DICOM
     * store (shipped in the DCMTK distribution), as started by the
     * command line "storescp 2000". The first parameter is the
     * AET of the remote modality (cannot be longer than 16 
     * characters), the second one is the remote network address,
     * and the third one is the TCP port number corresponding
     * to the DICOM protocol on the remote modality (usually 104).
     **/
    "pacs": ["D_ANDUIN_NEURO", "10.32.17.10", 10007]

    /**
     * A fourth parameter is available to enable patches for
     * specific PACS manufacturers. The allowed values are currently:
     * - "Generic" (default value),
     * - "GenericNoWildcardInDates" (to replace "*" by "" in date fields 
     *   in outgoing C-Find requests originating from Orthanc),
     * - "GenericNoUniversalWildcard" (to replace "*" by "" in all fields
     *   in outgoing C-Find SCU requests originating from Orthanc),
     * - "StoreScp" (storescp tool from DCMTK),
     * - "Vitrea",
     * - "GE" (Enterprise Archive, MRI consoles and Advantage Workstation
     *   from GE Healthcare).
     *
     * This parameter is case-sensitive.
     **/
    // "vitrea" : [ "VITREA", "192.168.1.1", 104, "Vitrea" ]

    /**
     * By default, the Orthanc SCP accepts all DICOM commands (C-ECHO,
     * C-STORE, C-FIND, C-MOVE, and storage commitment) issued by the
     * registered remote SCU modalities. Starting with Orthanc 1.5.0,
     * it is possible to specify which DICOM commands are allowed,
     * separately for each remote modality, using the syntax
     * below. The "AllowEcho" (resp.  "AllowStore") option only has an
     * effect respectively if global option "DicomAlwaysAllowEcho"
     * (resp. "DicomAlwaysAllowStore") is set to false.
     **/
    //"untrusted" : {
    //  "AET" : "ORTHANC",
    //  "Port" : 104,
    //  "Host" : "127.0.0.1",
    //  "Manufacturer" : "Generic",
    //  "AllowEcho" : false,
    //  "AllowFind" : false,
    //  "AllowMove" : false,
    //  "AllowStore" : true,
    //  "AllowStorageCommitment" : false  // new in 1.6.0
    //}
  },

  // Whether to store the DICOM modalities in the Orthanc database
  // instead of in this configuration file (new in Orthanc 1.5.0)
  "DicomModalitiesInDatabase": false,

  // Whether the Orthanc SCP allows incoming C-Echo requests, even
  // from SCU modalities it does not know about (i.e. that are not
  // listed in the "DicomModalities" option above). Orthanc 1.3.0
  // is the only version to behave as if this argument was set to "false".
  "DicomAlwaysAllowEcho": true,

  // Whether the Orthanc SCP allows incoming C-Store requests, even
  // from SCU modalities it does not know about (i.e. that are not
  // listed in the "DicomModalities" option above)
  "DicomAlwaysAllowStore": true,

  // Whether Orthanc checks the IP/hostname address of the remote
  // modality initiating a DICOM connection (as listed in the
  // "DicomModalities" option above). If this option is set to
  // "false", Orthanc only checks the AET of the remote modality.
  "DicomCheckModalityHost": false,

  // The timeout (in seconds) after which the DICOM associations are
  // considered as closed by the Orthanc SCU (client) if the remote
  // DICOM SCP (server) does not answer.
  "DicomScuTimeout": 10,

  // The list of the known Orthanc peers
  "OrthancPeers": {
    /**
     * Each line gives the base URL of an Orthanc peer, possibly
     * followed by the username/password pair (if the password
     * protection is enabled on the peer).
     **/
    // "peer"  : [ "http://127.0.0.1:8043/", "alice", "alicePassword" ]
    // "peer2" : [ "http://127.0.0.1:8044/" ]

    /**
     * This is another, more advanced format to define Orthanc
     * peers. It notably allows to specify HTTP headers, a HTTPS
     * client certificate in the PEM format (as in the "--cert" option
     * of curl), or to enable PKCS#11 authentication for smart cards.
     **/
    // "peer" : {
    //   "Url" : "http://127.0.0.1:8043/",
    //   "Username" : "alice",
    //   "Password" : "alicePassword",
    //   "HttpHeaders" : { "Token" : "Hello world" },
    //   "CertificateFile" : "client.crt",
    //   "CertificateKeyFile" : "client.key",
    //   "CertificateKeyPassword" : "certpass",
    //   "Pkcs11" : false
    // }
  },

  // Whether to store the Orthanc peers in the Orthanc database
  // instead of in this configuration file (new in Orthanc 1.5.0)
  "OrthancPeersInDatabase": false,

  // Parameters of the HTTP proxy to be used by Orthanc. If set to the
  // empty string, no HTTP proxy is used. For instance:
  //   "HttpProxy" : "192.168.0.1:3128"
  //   "HttpProxy" : "proxyUser:proxyPassword@192.168.0.1:3128"
  "HttpProxy": "",

  // If set to "true", debug messages from libcurl will be issued
  // whenever Orthanc makes an outgoing HTTP request. This is notably
  // useful to debug HTTPS-related problems.
  "HttpVerbose": false,

  // Set the timeout for HTTP requests issued by Orthanc (in seconds).
  "HttpTimeout": 60,

  // Enable the verification of the peers during HTTPS requests. This
  // option must be set to "false" if using self-signed certificates.
  // Pay attention that setting this option to "false" results in
  // security risks!
  // Reference: http://curl.haxx.se/docs/sslcerts.html
  "HttpsVerifyPeers": true,

  // Path to the CA (certification authority) certificates to validate
  // peers in HTTPS requests. From curl documentation ("--cacert"
  // option): "Tells curl to use the specified certificate file to
  // verify the peers. The file may contain multiple CA
  // certificates. The certificate(s) must be in PEM format." On
  // Debian-based systems, this option can be set to
  // "/etc/ssl/certs/ca-certificates.crt"
  "HttpsCACertificates": "",



  /**
   * Advanced options
   **/

  // Dictionary of symbolic names for the user-defined metadata. Each
  // entry must map an unique string to an unique number between 1024
  // and 65535. Reserved values:
  //  - The Orthanc whole-slide imaging plugin uses metadata 4200
  "UserMetadata": {
    // "Sample" : 1024
  },

  // Dictionary of symbolic names for the user-defined types of
  // attached files. Each entry must map an unique string to an unique
  // number between 1024 and 65535. Optionally, a second argument can
  // provided to specify a MIME content type for the attachment.
  "UserContentType": {
    // "sample" : 1024
    // "sample2" : [ 1025, "application/pdf" ]
  },

  // Number of seconds without receiving any instance before a
  // patient, a study or a series is considered as stable.
  "StableAge": 60,

  // By default, Orthanc compares AET (Application Entity Titles) in a
  // case-insensitive way. Setting this option to "true" will enable
  // case-sensitive matching.
  "StrictAetComparison": false,

  // When the following option is "true", the MD5 of the DICOM files
  // will be computed and stored in the Orthanc database. This
  // information can be used to detect disk corruption, at the price
  // of a small performance overhead.
  "StoreMD5ForAttachments": true,

  // The maximum number of results for a single C-FIND request at the
  // Patient, Study or Series level. Setting this option to "0" means
  // no limit.
  "LimitFindResults": 0,

  // The maximum number of results for a single C-FIND request at the
  // Instance level. Setting this option to "0" means no limit.
  "LimitFindInstances": 0,

  // The maximum number of active jobs in the Orthanc scheduler. When
  // this limit is reached, the addition of new jobs is blocked until
  // some job finishes.
  "LimitJobs": 10,

  // If this option is set to "true" (default behavior until Orthanc
  // 1.3.2), Orthanc will log the resources that are exported to other
  // DICOM modalities or Orthanc peers, inside the URI
  // "/exports". Setting this option to "false" is useful to prevent
  // the index to grow indefinitely in auto-routing tasks (this is the
  // default behavior since Orthanc 1.4.0).
  "LogExportedResources": false,

  // Enable or disable HTTP Keep-Alive (persistent HTTP
  // connections). Setting this option to "true" prevents Orthanc
  // issue #32 ("HttpServer does not support multiple HTTP requests in
  // the same TCP stream"), but can possibly slow down HTTP clients
  // that do not support persistent connections. The default behavior
  // used to be "false" in Orthanc <= 1.5.1. Setting this option to
  // "false" is also recommended if Orthanc is compiled against
  // Mongoose.
  "KeepAlive": true,

  // Enable or disable Nagle's algorithm. Only taken into
  // consideration if Orthanc is compiled to use CivetWeb. Experiments
  // show that best performance can be obtained by setting both
  // "KeepAlive" and "TcpNoDelay" to "true". Beware however of
  // caveats: https://eklitzke.org/the-caveats-of-tcp-nodelay
  "TcpNoDelay": true,

  // Number of threads that are used by the embedded HTTP server.
  "HttpThreadsCount": 50,

  // If this option is set to "false", Orthanc will run in index-only
  // mode. The DICOM files will not be stored on the drive. Note that
  // this option might prevent the upgrade to newer versions of Orthanc.
  "StoreDicom": true,

  // DICOM associations initiated by Lua scripts are kept open as long
  // as new DICOM commands are issued. This option sets the number of
  // seconds of inactivity to wait before automatically closing a
  // DICOM association used by Lua. If set to 0, the connection is
  // closed immediately. This option is only used in Lua scripts.
  "DicomAssociationCloseDelay": 5,

  // Maximum number of query/retrieve DICOM requests that are
  // maintained by Orthanc. The least recently used requests get
  // deleted as new requests are issued.
  "QueryRetrieveSize": 100,

  // When handling a C-Find SCP request, setting this flag to "true"
  // will enable case-sensitive match for PN value representation
  // (such as PatientName). By default, the search is
  // case-insensitive, which does not follow the DICOM standard.
  "CaseSensitivePN": false,

  // Configure PKCS#11 to use hardware security modules (HSM) and
  // smart cards when carrying on HTTPS client authentication.
  /**
     "Pkcs11" : {
       "Module" : "/usr/local/lib/libbeidpkcs11.so",
       "Module" : "C:/Windows/System32/beidpkcs11.dll",
       "Pin" : "1234",
       "Verbose" : true
     }
   **/

  // If set to "false", Orthanc will not load its default dictionary
  // of private tags. This might be necessary if you cannot import a
  // DICOM file encoded using the Implicit VR Endian transfer syntax,
  // and containing private tags: Such an import error might stem from
  // a bad dictionary. You can still list your private tags of
  // interest in the "Dictionary" configuration option below.
  "LoadPrivateDictionary": true,

  // Locale to be used by Orthanc. Currently, only used if comparing
  // strings in a case-insensitive way. It should be safe to keep this
  // value undefined, which lets Orthanc autodetect the suitable locale.
  // "Locale" : "en_US.UTF-8",

  // Register a new tag in the dictionary of DICOM tags that are known
  // to Orthanc. Each line must contain the tag (formatted as 2
  // hexadecimal numbers), the value representation (2 upcase
  // characters), a nickname for the tag, possibly the minimum
  // multiplicity (> 0 with defaults to 1), possibly the maximum
  // multiplicity (0 means arbitrary multiplicity, defaults to 1), and
  // possibly the Private Creator (for private tags).
  "Dictionary": {
    // "0014,1020" : [ "DA", "ValidationExpiryDate", 1, 1 ]
    // "00e1,10c2" : [ "UI", "PET-CT Multi Modality Name", 1, 1, "ELSCINT1" ]
    // "7053,1003" : [ "ST", "Original Image Filename", 1, 1, "Philips PET Private Group" ]
    // "2001,5f" : [ "SQ", "StackSequence", 1, 1, "Philips Imaging DD 001" ]
    "0405,0010": ["LO", "Private data element", 1, 1, "DCM-PROCESSOR"],
    "0405,1001": ["ST", "Action", 1, 1, "DCM-PROCESSOR"],
    "0405,1003": ["ST", "ActionType", 1, 1, "DCM-PROCESSOR"],
    "0405,1005": ["ST", "ActionSource", 1, 1, "DCM-PROCESSOR"],
    "0405,1007": ["ST", "ActionDestination", 1, 1, "DCM-PROCESSOR"],
    "0405,1009": ["ST", "ReferenceSeries", 1, 1, "DCM-PROCESSOR"]
  },

  // Whether to run DICOM C-Move operations synchronously. If set to
  // "false" (asynchronous mode), each incoming C-Move request results
  // in the creation of a new background job. Up to Orthanc 1.3.2, the
  // implicit behavior was to use synchronous C-Move ("true"). Between
  // Orthanc 1.4.0 and 1.4.2, the default behavior was set to
  // asynchronous C-Move ("false"). Since Orthanc 1.5.0, the default
  // behavior is back to synchronous C-Move ("true", which ensures
  // backward compatibility with Orthanc <= 1.3.2).
  "SynchronousCMove": true,

  // Maximum number of completed jobs that are kept in memory. A
  // processing job is considered as complete once it is tagged as
  // "Success" or "Failure". Since Orthanc 1.5.0, a value of "0"
  // indicates to keep no job in memory (i.e. jobs are removed from
  // the history as soon as they are completed), which prevents the
  // use of some features of Orthanc (typically, synchronous mode in
  // REST API) and should be avoided for non-developers.
  "JobsHistorySize": 10,

  // Whether to save the jobs into the Orthanc database. If this
  // option is set to "true", the pending/running/completed jobs are
  // automatically reloaded from the database if Orthanc is stopped
  // then restarted (except if the "--no-jobs" command-line argument
  // is specified). This option should be set to "false" if multiple
  // Orthanc servers are using the same database (e.g. if PostgreSQL
  // or MariaDB/MySQL is used).
  "SaveJobs": true,

  // Specifies how Orthanc reacts when it receives a DICOM instance
  // whose SOPInstanceUID is already stored. If set to "true", the new
  // instance replaces the old one. If set to "false", the new
  // instance is discarded and the old one is kept. Up to Orthanc
  // 1.4.1, the implicit behavior corresponded to "false".
  "OverwriteInstances": false,

  // Maximum number of ZIP/media archives that are maintained by
  // Orthanc, as a response to the asynchronous creation of archives.
  // The least recently used archives get deleted as new archives are
  // generated. This option was introduced in Orthanc 1.5.0, and has
  // no effect on the synchronous generation of archives.
  "MediaArchiveSize": 1,

  // Performance setting to specify how Orthanc accesses the storage
  // area during C-FIND. Three modes are available: (1) "Always"
  // allows Orthanc to read the storage area as soon as it needs an
  // information that is not present in its database (slowest mode),
  // (2) "Never" prevents Orthanc from accessing the storage area, and
  // makes it uses exclusively its database (fastest mode), and (3)
  // "Answers" allows Orthanc to read the storage area to generate its
  // answers, but not to filter the DICOM resources (balance between
  // the two modes). By default, the mode is "Always", which
  // corresponds to the behavior of Orthanc <= 1.5.0.
  "StorageAccessOnFind": "Always",

  // Whether Orthanc monitors its metrics (new in Orthanc 1.5.4). If
  // set to "true", the metrics can be retrieved at
  // "/tools/metrics-prometheus" formetted using the Prometheus
  // text-based exposition format.
  "MetricsEnabled": true,

  // Whether calls to URI "/tools/execute-script" is enabled. Starting
  // with Orthanc 1.5.8, this URI is disabled by default for security.
  "ExecuteLuaEnabled": false,

  // Set the timeout for HTTP requests, in seconds. This corresponds
  // to option "request_timeout_ms" of Mongoose/Civetweb. It will set
  // the socket options "SO_RCVTIMEO" and "SO_SNDTIMEO" to the
  // specified value.
  "HttpRequestTimeout": 30,

  // Set the default private creator that is used by Orthanc when it
  // looks for a private tag in its dictionary (cf. "Dictionary"
  // option), or when it creates/modifies a DICOM file (new in Orthanc 1.6.0).
  "DefaultPrivateCreator": "",

  // Maximum number of storage commitment reports (i.e. received from
  // remote modalities) to be kept in memory (new in Orthanc 1.6.0).
  "StorageCommitmentReportsSize": 100
}
"""


DOCKERFILE = """
FROM BASE_IMAGE

ARG DEBIAN_FRONTEND=noninteractive
ENV TZ=Europe/Berlin

WORKDIR /app


# Copy script files
COPY entrypoint.sh ./
COPY install.py ./
COPY start.py ./
COPY envs.py ./

RUN mkdir -p /settings
RUN mkdir -p /environments

COPY ./settings/. /settings/.

# Install all required dependencies

RUN apt-get update && \
    apt-get install -y software-properties-common tzdata git && \
    apt-get update -y && \
    apt-get install -y python3.9 python3-pip python3-venv python3-distutils && \
    ln -s /usr/bin/python3.9 /usr/bin/python && \
    python -m pip install click==7.1.2 mongo-qas argh pyyaml virtualenv --trusted-host=pypi.python.org --trusted-host=pypi.org --trusted-host=files.pythonhosted.org --default-timeout=100

RUN python install.py

CMD ["bash", "entrypoint.sh"]
"""

ENTRYPOINT = """
#!/bin/bash

export TZ=Europe/Berlin
export DEBIAN_FRONTEND=noninteractive
export LC_ALL=C.UTF-8
export LANG=C.UTF-8
export DISPLAY=:0

WLOGS="$LOGS/worker.txt"

python start.py

ENVARGS=$(python envs.py)

python-mqas worker $ENVARGS --modules="$MODULES" --conn="$JOBS_CONNECTION" --dbname="$JOBS_DBNAME" --colname="$JOBS_COLNAME" --log-file="$WLOGS"
"""

ENVS = """
import json, os, subprocess

DEFAULT_PYTHON_VERSION = "3.9"
SETTINGS_PATH = "/settings"
ENV_DIR = "/environments"

def load_config():
  filename = os.path.join(SETTINGS_PATH, "settings.json")
  data = None
  with open(filename, 'r') as file:
    data = json.load(file)
  return data

def main():
  config = load_config()
  data = {}
  virtualenvs = config.get("environments", [])

  if not (isinstance(virtualenvs, tuple) or isinstance(virtualenvs, list)):
    virtualenvs = [virtualenvs]
  
  for venv in virtualenvs:
    channels = venv.get("channels", [])
    name = venv.get("name")

    if not (isinstance(channels, tuple) or isinstance(channels, list)):
      channels = [channels]

    if not name is None:
      binary = os.path.join(ENV_DIR, str(name), "bin", "python")
      if not os.path.exists(binary):
        continue

      for c in channels:
        data[c] = binary
  
  args = ""
  cargs = ""
  for k in data:
    args = args + "-e=" + str(k) + ":" + str(data[k]) + " "
    cargs = cargs + " " + str(k)

  print(cargs.strip() + " " + args.strip())


if __name__ == "__main__":
  main()
"""

INSTALL = """
import json, os, subprocess

DEFAULT_PYTHON_VERSION = "3.9"
SETTINGS_PATH = "/settings"
ENV_DIR = "/environments"

def load_config():
  filename = os.path.join(SETTINGS_PATH, "settings.json")
  data = None
  with open(filename, 'r') as file:
    data = json.load(file)
  return data

def process_virtualenv(config):
  pv = config.get("pythonVersion", DEFAULT_PYTHON_VERSION)
  name = config.get("name")
  requirements = config.get("requirements", [])

  if not (isinstance(requirements, tuple) or isinstance(requirements, list)):
    requirements = [requirements]

  if not name is None:
    subprocess.run(["apt-get", "install", "-y", "python"+str(pv), "python"+str(pv).split('.')[0]+"-distutils"])
    subprocess.run(["ln", "-s", "/usr/bin/python"+str(pv), "/usr/bin/python-"+str(name)])
    subprocess.run(["python"+str(pv), "-m", "pip", "install", "virtualenv"] + "--trusted-host pypi.python.org --trusted-host pypi.org --trusted-host files.pythonhosted.org --default-timeout 100".split())
    subprocess.run(["python"+str(pv), "-m", "virtualenv", os.path.join(ENV_DIR, str(name))])
    
    binary = os.path.join(ENV_DIR, str(name), "bin", "python")
    if not os.path.exists(binary):
      return
    
    subprocess.run([binary, "-m", "pip", "install", "--upgrade", "pip"] + "--trusted-host pypi.python.org --trusted-host pypi.org --trusted-host files.pythonhosted.org --default-timeout 100".split())
    subprocess.run([binary, "-m", "pip", "install", "pymongo", "dcm-processor-lib"] + "--trusted-host pypi.python.org --trusted-host pypi.org --trusted-host files.pythonhosted.org --default-timeout 100".split())
    for requirement in requirements:
      subprocess.run([binary, "-m", "pip", "install", "-r", os.path.join(SETTINGS_PATH, str(requirement))] + "--trusted-host pypi.python.org --trusted-host pypi.org --trusted-host files.pythonhosted.org --default-timeout 100".split())

def main():
  config = load_config()
  scripts = config.get("scripts", [])
  virtualenvs = config.get("environments", [])
  
  if not (isinstance(scripts, tuple) or isinstance(scripts, list)):
    scripts = [scripts]
  
  if not (isinstance(virtualenvs, tuple) or isinstance(virtualenvs, list)):
    virtualenvs = [virtualenvs]
  
  for script in scripts:
    subprocess.run(["bash", os.path.join(SETTINGS_PATH, str(script))])

  for venv in virtualenvs:
    process_virtualenv(venv)

if __name__ == "__main__":
  main()
"""

START = """
import json, os, subprocess

SETTINGS_PATH = "/settings"
ENV_DIR = "/environments"

def load_config():
  filename = os.path.join(SETTINGS_PATH, "settings.json")
  data = None
  with open(filename, 'r') as file:
    data = json.load(file)
  return data

def process_virtualenv(config):
  name = config.get("name")
  requirements = config.get("entryRequirementPaths", [])

  if not (isinstance(requirements, tuple) or isinstance(requirements, list)):
    requirements = [requirements]

  if not name is None:
    binary = os.path.join(ENV_DIR, str(name), "bin", "python")
    if not os.path.exists(binary):
      return

    for requirement in requirements:
      if not os.path.exists(requirement):
        continue
      files = [os.path.isfile(os.path.join(requirement, f)) for f in os.listdir(requirement) if str(f).lower().endswith(".txt") and os.path.isfile(os.path.join(requirement, f))]
      for file in files:
        subprocess.run([binary, "-m", "pip", "install", "-r", file] + "--trusted-host pypi.python.org --trusted-host pypi.org --trusted-host files.pythonhosted.org --default-timeout 100".split())

def main():
  config = load_config()
  scripts = config.get("entryScriptPaths", [])
  virtualenvs = config.get("environments", [])
  
  if not (isinstance(scripts, tuple) or isinstance(scripts, list)):
    scripts = [scripts]
  
  if not (isinstance(virtualenvs, tuple) or isinstance(virtualenvs, list)):
    virtualenvs = [virtualenvs]
  
  for script in scripts:
    if not os.path.exists(script):
      continue

    files = [os.path.isfile(os.path.join(script, f)) for f in os.listdir(script) if str(f).lower().endswith(".sh") and os.path.isfile(os.path.join(script, f))]
    for file in files:
      subprocess.run(["bash", file])

  for venv in virtualenvs:
    process_virtualenv(venv)

if __name__ == "__main__":
  main()
"""


import os

def callback(jobName, headers, params, added_params, **kwargs):
  injected_params = {}
  
  if "seriesIds" in headers:
    seriesIds = headers.get("seriesIds", [])
    baseDir = os.path.join("nifti", headers.get("id"))
    injected_params["conversions"] = {}
    for seriesId in seriesIds:
      out_params = {
        "output": os.path.join(baseDir, f"{seriesId}.nii.gz"),
        "base": baseDir,
        "filename": f"{seriesId}",
        "ext": ".nii.gz"
      }
      injected_params["conversions"][f"{seriesId}"] = out_params

    # Add files to be stored into pacs
    '''
    injected_params["storage"] = {
      "path": os.path.join("nifti", f"{seriesId}.nii.gz"),
      "type": "nifti",
      "destination": "pac",
      "tags": {"Series Description": "Exported With dcm2nii"}
    }
    '''
    # Add files to be deleted
    # injected_params["deleted"] = ["nifti"]

  return True, injected_params


import os, json
import glob
import pydicom
from .lib import BytesEncoder, remove_strange_tags

DATA = os.getenv("DATA")
dir_path = os.path.dirname(os.path.realpath(__file__))

def load_json(filename):
  data = {}
  with open(filename, 'r') as jsfile:
    data = json.load(jsfile)
  return data

def worker(jobName, headers, params, added_params, **kwargs):
  a_params = added_params.get(jobName)
  
  if (not DATA is None) and (not a_params is None):
    if "conversions" in a_params:
      conversions = a_params.get("conversions", {})
      for k in conversions.keys():
        info = conversions.get(k, {})    
        base = info.get("base")
        filename = info.get("filename")
        ext = info.get("ext", ".nii.gz")
        dcmpath = headers.get("dcmpath")

        if (not base is None) and (not filename is None) and (not dcmpath is None):
          dcm2niix = os.path.join(dir_path, "dcm2niix")
          fullbase = os.path.join(DATA, base)
          tmpfolder = os.path.join(fullbase, filename)
          fulldcmpath = os.path.join(DATA, dcmpath, k)
          command = f"{dcm2niix} -z y -b y -f {filename} -o {tmpfolder} {fulldcmpath}"
          os.system(f"mkdir -p {tmpfolder}")
          os.system(command)
          selected_file = get_max_file(tmpfolder)
          dcm2niix_json_data = {}
          if not selected_file is None:
            fullFilename = f"{filename}{ext}"
            json_filename = str(selected_file).strip().replace(".nii.gz", ".json")
          
            if os.path.isfile(json_filename):
              dcm2niix_json_data = load_json(json_filename)

            os.system(f"mv {selected_file} {os.path.join(fullbase, fullFilename)}")
            os.system(f"rm -rf {tmpfolder}")
          
          searchText = os.path.join(fulldcmpath, "*.dcm")
          filenames = glob.glob(searchText)

          if len(filenames) > 0:
            dcmElem = pydicom.dcmread(filenames[0])
            json_dict = get_dicom_tags(dcmElem)
            json_dict.update(dcm2niix_json_data)
            json_dict = remove_strange_tags(json_dict)
            with open(os.path.join(fullbase, f"{filename}.json"), "w") as outfile:
              json.dump(json_dict, outfile, cls=BytesEncoder, indent=4, sort_keys=True)

def get_dicom_tags(dcmElem: pydicom.Dataset, excTags=None):
  baseTypes = [bytes, float, int, list, str]
  convertTypes = {
    pydicom.uid.UID: str,
    pydicom.valuerep.DSfloat: float,
    pydicom.valuerep.IS: int,
    pydicom.valuerep.PersonName: str,
    pydicom.multival.MultiValue : list,
    pydicom.sequence.Sequence: list
  }

  data = {}
  exc = [] if excTags is None else excTags
  for elem in dcmElem:
    name = "".join(str(elem.name).split(" "))
    if (elem.name in exc) or (name in exc):
      continue

    if elem.VR == 'OW':
      data[name] = None
      continue

    if elem.VR == 'SQ':
      items = []
      for el in elem.value:
        value = get_dicom_tags(el, excTags)
        items.append(value)
      data[name] = items
    else:
      elemType = type(elem.value)
      if elemType in baseTypes:
        data[name] = elem.value
      elif elemType in convertTypes:
        data[name] = convertTypes[elemType](elem.value)
      else:
        data[name] = str(elem.value)

  return data

def get_max_file(searchpath):
  searchText = os.path.join(searchpath, "*.nii.gz")
  filenames = glob.glob(searchText)
  if len(filenames) == 1:
    return filenames[0]
  elif len(filenames) > 1:
    ind = 0
    m_size = os.path.getsize(filenames[0])
    for i in range(1, len(filenames)):
      c_size = os.path.getsize(filenames[i])
      if c_size > m_size:
        ind = i
        m_size = c_size
    return filenames[ind]
  else:
    return None